/* -*- tab-width: 4; -*- */
/*------------------------------------------------------------------*
 * Filename:		libnbt_ns.c
 *------------------------------------------------------------------*/

#include <stdio.h>

#ifdef WIN32
#include <WinSock.h>
#include <errno.h>

#else // WIN32
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/time.h>

#endif // WIN32

#include "libnbt.h"
#include "libnbt_ns.h"
#include "libnbt_private.h"
#include "libnbt_glob.h"
#include "libnbt_ns_in.h"
#include "libnbt_mes.h"
#include "libnbt_util.h"
#include "libnbt_util_private.h"
#include "libnbt_err.h"
#include "libnbt_debug.h"

// private function def
static unsigned char getGroupValueFromMacro(unsigned char macro);
static unsigned char getGroupMacroFromValue(unsigned char value);
#if 0
static unsigned char getRcodeValueFromMacro(unsigned char macro);
#endif
static unsigned char getRcodeMacroFromValue(unsigned char value);
static unsigned char getONTMacroFromValue(unsigned char value);
static unsigned char getStateMacroFromValue(unsigned char value);
static int getNBTNodeNameFromEntry(NBTNodeName* dst, const stNameServiceNodeNameEntry* src);
static void getNBTNodeStatisticsFromServiceStatistics(NBTNodeStatistics* dst, const stNameServiceStatistics* src);


// addname
static int addname_broadcast(NBTCTBL* ct, const NBTNamePack* np, int mestype, unsigned char* rcode,
							 NBTForceSettingInfo* fcinfo);
static int addname_unicast(NBTCTBL* ct, const NBTNamePack* np, unsigned long winsnaddr,
						   unsigned long* refresh, int mestype, unsigned char* rcode,
						   NBTForceSettingInfo* fcinfo);
static int addname_b_node(NBTCTBL* ct, const NBTNamePack* np, unsigned long* refresh, int mestype, unsigned char* rcode,
						  NBTForceSettingInfo* fcinfo);
static int addname_p_node(NBTCTBL* ct, const NBTNamePack* np, unsigned long winsnaddr,
						  unsigned long* refresh, int mestype, unsigned char* rcode,
						  NBTForceSettingInfo* fcinfo);
static int addname_m_node(NBTCTBL* ct, const NBTNamePack* np, unsigned long winsnaddr,
						  unsigned long* refresh, int mestype, unsigned char* rcode,
						  NBTForceSettingInfo* fcinfo);
static int addname_h_node(NBTCTBL* ct, const NBTNamePack* np, unsigned long winsnaddr,
						  unsigned long* refresh, int mestype, unsigned char* rcode,
						  NBTForceSettingInfo* fcinfo);

// findname
static int findname_broadcast(NBTCTBL* ct, const NBTNamePack* np, NBTAddrPack* ap, int num, int* findnum, NBTForceSettingInfo* fcinfo);
static int findname_unicast(NBTCTBL* ct, const NBTNamePack* np, unsigned long winsnaddr,
							NBTAddrPack* ap, int num, int* findnum, NBTForceSettingInfo* fcinfo);
static int findname_b_node(NBTCTBL* ct, const NBTNamePack* np, NBTAddrPack* ap, int num, int* findnum, NBTForceSettingInfo* fcinfo);
static int findname_p_node(NBTCTBL* ct, const NBTNamePack* np, unsigned long winsnaddr,
						   NBTAddrPack* ap, int num, int* findnum, NBTForceSettingInfo* fcinfo);
static int findname_m_node(NBTCTBL* ct, const NBTNamePack* np, unsigned long winsnaddr,
						   NBTAddrPack* ap, int num, int* findnum, NBTForceSettingInfo* fcinfo);
static int findname_h_node(NBTCTBL* ct, const NBTNamePack* np, unsigned long winsnaddr,
						   NBTAddrPack* ap, int num, int* findnum, NBTForceSettingInfo* fcinfo);

// deletename
static int deletename_broadcast(NBTCTBL* ct, const NBTNamePack* np, NBTForceSettingInfo* fcinfo);
static int deletename_unicast(NBTCTBL* ct, const NBTNamePack* np, unsigned long winsnaddr,
							  NBTForceSettingInfo* fcinfo);
static int deletename_b_node(NBTCTBL* ct, const NBTNamePack* np, NBTForceSettingInfo* fcinfo);
static int deletename_p_node(NBTCTBL* ct, const NBTNamePack* np, unsigned long winsnaddr, NBTForceSettingInfo* fcinfo);
static int deletename_m_node(NBTCTBL* ct, const NBTNamePack* np, unsigned long winsnaddr, NBTForceSettingInfo* fcinfo);
static int deletename_h_node(NBTCTBL* ct, const NBTNamePack* np, unsigned long winsnaddr, NBTForceSettingInfo* fcinfo);


/*
------------------------------------------------------------------

------------------------------------------------------------------
*/
static unsigned char getGroupValueFromMacro(unsigned char macro)
{
	unsigned char val = -1;

	switch(macro){
	case NBT_G_NONE:	break;
	case NBT_G_UNIQUE:	val = VAL_G_UNIQUE; break;
	case NBT_G_GROUP:	val = VAL_G_GROUP; break;
	default:
		D_PRINTF(("[libnbt] getGroupValueToMacro: unsupport GroupMacro\n"));
		break;
	}
	return val;
}

/*
------------------------------------------------------------------

------------------------------------------------------------------
*/
static unsigned char getGroupMacroFromValue(unsigned char value)
{
	unsigned char macro = -1;

	switch(value){
	case VAL_G_UNIQUE:	macro = NBT_G_UNIQUE; break;
	case VAL_G_GROUP:	macro = NBT_G_GROUP; break;
	default:
		D_PRINTF(("[libnbt] getGroupMacroToValue: unsupport GroupValue %u\n", value));
		break;
	}
	return macro;
}


/*
------------------------------------------------------------------

------------------------------------------------------------------
*/
#if 0
static unsigned char getRcodeValueFromMacro(unsigned char macro)
{
	unsigned char val = -1;

	switch(macro){
	case NBT_RC_SUCCESS:	val = VAL_RC_SUCCESS;	break;
	case NBT_RC_FMT_ERR:	val = VAL_RC_FMT_ERR;	break;
	case NBT_RC_SRV_ERR:	val = VAL_RC_SRV_ERR;	break;
	case NBT_RC_NAM_ERR:	val = VAL_RC_NAM_ERR;	break;
	case NBT_RC_IMP_ERR:	val = VAL_RC_IMP_ERR;	break;
	case NBT_RC_RFS_ERR:	val = VAL_RC_RFS_ERR;	break;
	case NBT_RC_ACT_ERR:	val = VAL_RC_ACT_ERR;	break;
	case NBT_RC_CFT_ERR:	val = VAL_RC_CFT_ERR;	break;
	default:
		D_PRINTF(("[libnbt] getRcodeValueFromMacro: unsupport RcodeMacro %u\n", macro));
		break;
	}

	return val;
}
#endif

/*
------------------------------------------------------------------

------------------------------------------------------------------
*/
static unsigned char getRcodeMacroFromValue(unsigned char value)
{
	unsigned char macro = -1;

	switch(value){
	case VAL_RC_SUCCESS:	macro = NBT_RC_SUCCESS;	break;
	case VAL_RC_FMT_ERR:	macro = NBT_RC_FMT_ERR;	break;
	case VAL_RC_SRV_ERR:	macro = NBT_RC_SRV_ERR;	break;
	case VAL_RC_NAM_ERR:	macro = NBT_RC_NAM_ERR;	break;
	case VAL_RC_IMP_ERR:	macro = NBT_RC_IMP_ERR;	break;
	case VAL_RC_RFS_ERR:	macro = NBT_RC_RFS_ERR;	break;
	case VAL_RC_ACT_ERR:	macro = NBT_RC_ACT_ERR;	break;
	case VAL_RC_CFT_ERR:	macro = NBT_RC_CFT_ERR;	break;
	default:
		D_PRINTF(("[libnbt] getRcodeMacroFromValue: unsupport RcodeValue %u\n", value));
		break;
	}

	return macro;
}


/*
------------------------------------------------------------------

------------------------------------------------------------------
*/
static unsigned char getONTMacroFromValue(unsigned char value)
{
	unsigned char macro = -1;
	
	switch(value){
	case VAL_ONT_B:	macro = NBT_ONT_B; break;
	case VAL_ONT_P:	macro = NBT_ONT_P; break;
	case VAL_ONT_M:	macro = NBT_ONT_M; break;
	case VAL_ONT_H:	macro = NBT_ONT_H; break;
	default:
		D_PRINTF(("[libnbt] getONTMacroFromValue: unsupport ONTValue %u\n", value));
		break;
	}
	
	return macro;
}

/*
------------------------------------------------------------------

------------------------------------------------------------------
*/
static unsigned char getStateMacroFromValue(unsigned char value)
{
	unsigned char macro = 0;
	
	if(value & NS_NNE_STATE_DRG_ON){
		macro |= NBT_NODESTATE_DRG;
	}
	if(value & NS_NNE_STATE_CNF_ON){
		macro |= NBT_NODESTATE_CNF;
	}
	if(value & NS_NNE_STATE_ACT_ON){
		macro |= NBT_NODESTATE_ACT;
	}
	if(value & NS_NNE_STATE_PRM_ON){
		macro |= NBT_NODESTATE_PRM;
	}
	
	return macro;
}


/*
------------------------------------------------------------------
stNameServiceNodeNameEntry --> NBTNodeNameւ̕ϊ֐
------------------------------------------------------------------
*/
static int getNBTNodeNameFromEntry(NBTNodeName* dst, const stNameServiceNodeNameEntry* src)
{
	int res;
	
	memset(dst, 0, sizeof(NBTNodeName));
	res = nbt_devidename(dst->name, &dst->suffix, src->nbname);
	if(res < 0){
		return -1;
	}
	dst->group	= getGroupMacroFromValue(src->G);
	dst->ont	= getONTMacroFromValue(src->ONT);
	dst->state	= getStateMacroFromValue(src->state);
	
	return 0;
}


/*
------------------------------------------------------------------
stNameServiceStatistics --> NBTNodeStatisticsւ̕ϊ֐
------------------------------------------------------------------
*/
static void getNBTNodeStatisticsFromServiceStatistics(NBTNodeStatistics* dst, const stNameServiceStatistics* src)
{
	memcpy(dst->unit_id, src->UNIT_ID, sizeof dst->unit_id);
	dst->jumpers				= src->JUMPERS;
	dst->test_result			= src->TEST_RESULT;
	dst->version_number			= src->VERSION_NUMBER;
	dst->period_of_statistics	= src->PERIOD_OF_STATISTICS;
	dst->number_of_crcs	= src->NUMBER_OF_CRCs;
	dst->number_alignment_errors	= src->NUMBER_ALIGNMENT_ERRORS;
	dst->number_of_collisions	= src->NUMBER_OF_COLLISIONS;
	dst->number_send_aborts		= src->NUMBER_SEND_ABORTS;
	dst->number_good_sends		= src->NUMBER_GOOD_SENDS;
	dst->number_good_receives	= src->NUMBER_GOOD_RECEIVES;
	dst->number_retransmits		= src->NUMBER_RETRANSMITS;
	dst->number_no_resource_conditions	= src->NUMBER_NO_RESOURCE_CONDITIONS;
	dst->number_free_command_blocks	= src->NUMBER_FREE_COMMAND_BLOCKS;
	dst->total_number_command_blocks	= src->TOTAL_NUMBER_COMMAND_BLOCKS;
	dst->max_total_number_command_blocks	= src->MAX_TOTAL_NUMBER_COMMAND_BLOCKS;
	dst->number_pending_sessions	= src->NUMBER_PENDING_SESSIONS;
	dst->max_number_pending_sessions	= src->MAX_NUMBER_PENDING_SESSIONS;
	dst->max_total_sessions_possible	= src->MAX_TOTAL_SESSIONS_POSSIBLE;
	dst->session_data_packet_size	= src->SESSION_DATA_PACKET_SIZE;
}


////////////// addname

// -1: error
// 0: failure
// 1: success
/*
------------------------------------------------------------------
addname_broadcast()

u[hLXgŖOo^s

߂l:
	:
		LIBNBT_ERR_NOERR

	s:
		LIBNBT_ERR_INTERRUPTED
		LIBNBT_ERR_NAMEREG_FAILED ...rcodeɒlݒ肳邱Ƃ(rcode == -1: Ȃ)
		LIBNBT_ERR_OTHER_ERROR ... Ȃ̃G[(rcodeɒl͐ݒ肳Ȃ)

------------------------------------------------------------------
*/
static int addname_broadcast(NBTCTBL* ct, const NBTNamePack* np, int mestype, unsigned char* rcode,
							 NBTForceSettingInfo* fcinfo)
{
	int res;
	stRequestMessageInfo info;
	stNameServicePrimitiveRequest rinfo;
	unsigned char c;

	/* NameRegistrationpPbg쐬 */
	// create
	memset(&info, 0, sizeof info);
	info.fd 			= ct->fd;
	info.mestype 		= mestype;
	info.dstnaddr 		= ct->bcnaddr;
	info.dstnport 		= htons(NBT_PORT_NAME);
	info.retry 			= fcinfo? ((fcinfo->bcast_retry_count != -1)?
								fcinfo->bcast_retry_count: BCAST_REQ_RETRY_COUNT): BCAST_REQ_RETRY_COUNT;
	info.timeout		= fcinfo? ((fcinfo->bcast_retry_timeo != ~0)?
								fcinfo->bcast_retry_timeo: BCAST_REQ_RETRY_TIMEOUT): BCAST_REQ_RETRY_TIMEOUT;
	info.waitresponse	= 1;

	/* END-NODE-CHALLENGE RESPONSE  POSITIVE RESPONSE𖳎 */
	info.ignoreresponse	= NS_RESTYPE_ENODECHA | NS_RESTYPE_POSITIVE;
	info.rinfo[0] 		= (void*)&rinfo;
	info.interruptd		= fcinfo? ((fcinfo->interruptd != -1)? fcinfo->interruptd: -1): -1;
	info.prim.TranID 	= ct->tid;
	info.prim.TTL		= 0;

	nbt_padname(np->name, info.prim.QName, NBT_LENGTH_NETBIOSNAME, ' ', np->suffix);

	strcpy(info.prim.QScopeId, ct->scopeid? ct->scopeid: GSetting.scopeid);	// scopeid

	info.prim.NbAddress		= (np->naddr == 0)? ct->srcnaddr: np->naddr;
	info.prim.ONT			= GSetting.ont;
	info.prim.G				= getGroupValueFromMacro(np->group);
	info.prim.nmflags		= NS_NSPR_NMFLAGS_B_ON | DEF_NMFLAGS_REQ_NAMEREG;

	/* pPbgM */
	res = send_RequestMessageRepeatAnyTimes(&info);
	ct->tid = info.prim.TranID;
	if(res != LIBNBT_ERR_NOERR){
		if(res == LIBNBT_ERR_INTERRUPTED){
			return res;
		}
		if(res != LIBNBT_ERR_NORESPONSE){
			D_PRINTF(("[libnbt] addname_broadcast: send_RequestMessageRepeatAnyTimes failed\n"));
			return LIBNBT_ERR_OTHER_ERR;
		}
	}
	
	/* 肩烌X|XB */
	if(res == LIBNBT_ERR_NORESPONSE){
		// no response
		// create
		/* XVvbZ[W쐬 */
		info.mestype = NBT_NS_MT_REQ_REF;
		info.waitresponse = 0;
		info.retry = 1;
		
		/* XVvbZ[W𑗐M */
		res = send_RequestMessageRepeatAnyTimes(&info);
		ct->tid = info.prim.TranID;
		if(res != LIBNBT_ERR_NOERR){
			if(res == LIBNBT_ERR_INTERRUPTED){
				return res;
			}
			D_PRINTF(("[libnbt] addname_addgroup_b_node: send_RequestMessageRepeatAnyTimes failed.\n"));
			return LIBNBT_ERR_OTHER_ERR;
		}
		
		return LIBNBT_ERR_NOERR; // success
	}
	// response	
	/* 肩烌X|X */
	switch(info.retval){
	case NBT_NS_MT_RES_NEG_NAMEREG:
		/* rcodeɒl */
		c = getRcodeMacroFromValue(info.prim.Rcode);
		if(c == (unsigned char)-1){
			/* rcodes */
			return LIBNBT_ERR_OTHER_ERR;
		}
		*rcode = c;
		
		return LIBNBT_ERR_NAMEREG_FAILED;	// failure
	default:
		D_PRINTF(("[libnbt] addname_addgroup_b_node: send_RequestMessageRepeatAnyTimes no such response type.\n"));
		return LIBNBT_ERR_OTHER_ERR;
	}
	
	// never come here.
	return LIBNBT_ERR_OTHER_ERR;
}


/*
------------------------------------------------------------------
addname_unicast()

jLXgWINST[o[ɑ΂Oo^s

߂l:
	:	
		LIBNBT_ERR_NOERR

	s:
		LIBNBT_ERR_INVALID_PARAM
		LIBNBT_ERR_INTERRUPTED
		LIBNBT_ERR_NAMEREG_FAILED... rcodeɒlݒ肳(rcode == -1: Ȃ)
		LIBNBT_ERR_OTHER_ERROR

------------------------------------------------------------------
*/
static int addname_unicast(NBTCTBL* ct, const NBTNamePack* np, unsigned long winsnaddr,
						   unsigned long* refresh, int mestype, unsigned char* rcode,
						   NBTForceSettingInfo* fcinfo)
{
	stRequestMessageInfo info;
	stNameServicePrimitiveRequest rinfo;
	stNameServiceAddrEntry qdata;

	int res;
	int num;
	unsigned char c;

	*refresh = 0;

	if(winsnaddr == 0){
		return LIBNBT_ERR_INVALID_PARAM;
	}

	// create
	/* Oo^bZ[W쐬 */
	memset(&info, 0, sizeof info);
	info.fd 			= ct->fd;
	info.mestype 		= mestype;
	info.dstnaddr 		= winsnaddr;
	info.dstnport 		= htons(NBT_PORT_NAME);
	info.retry			= fcinfo? ((fcinfo->ucast_retry_count != -1)?
							fcinfo->ucast_retry_count: UCAST_REQ_RETRY_COUNT): UCAST_REQ_RETRY_COUNT;
	info.timeout		= fcinfo? ((fcinfo->ucast_retry_timeo != ~0)?
							fcinfo->ucast_retry_timeo: UCAST_REQ_RETRY_TIMEOUT): UCAST_REQ_RETRY_TIMEOUT;

	info.waitresponse	= 1;
	info.ignoreresponse	= 0;
	info.rinfo[0] 		= (void*)&rinfo;
	info.interruptd		= fcinfo? ((fcinfo->interruptd != -1)? fcinfo->interruptd: -1): -1;
	info.prim.TranID 	= ct->tid;
	info.prim.TTL		= 0;

	nbt_padname(np->name, info.prim.QName, NBT_LENGTH_NETBIOSNAME, ' ', np->suffix);

	strcpy(info.prim.QScopeId, ct->scopeid? ct->scopeid: GSetting.scopeid);	// scopeid

	info.prim.NbAddress		= (np->naddr == 0)? ct->srcnaddr: np->naddr;
	info.prim.ONT			= GSetting.ont;
	info.prim.G				= getGroupValueFromMacro(np->group);
	info.prim.nmflags		= DEF_NMFLAGS_REQ_NAMEREG;
	
	/* Oo^bZ[W𑗐M */
	res = send_RequestMessageRepeatAnyTimes(&info); // unicast
	ct->tid = info.prim.TranID;
	if(res != LIBNBT_ERR_NOERR){
		D_PRINTF(("[libnbt] addname_addgroup_p_node: send_RequestMessageRepeatAnyTimes failed\n"));
		if(res == LIBNBT_ERR_INTERRUPTED){
			return res;
		}
		return LIBNBT_ERR_OTHER_ERR;
	}
	
	/* X|X */
	// response	
	switch(info.retval){
	case NBT_NS_MT_RES_POS_NAMEREG:
		D_PRINTF(("[libnbt] addname_addgroup_p_node: INFO: adjust refresh time = %lu\n", rinfo.TTL));
		*refresh = rinfo.TTL;
		return LIBNBT_ERR_NOERR;	// success
		
	case NBT_NS_MT_RES_NEG_NAMEREG:
		/* read rcode */
		c = getGroupMacroFromValue(info.prim.Rcode);
		if(c == (unsigned char)-1){
			return LIBNBT_ERR_OTHER_ERR;
		}

		return LIBNBT_ERR_NAMEREG_FAILED;	// failure
		
	case NBT_NS_MT_RES_ENCREG:
		// create
		num = 1;
		
		info.mestype = NBT_NS_MT_REQ_NAMEQUERY;
		info.dstnaddr = rinfo.NbAddress;
		info.ignoreresponse = NS_RESTYPE_REDIRECT;
		info.rinfo[1] = (void*)&qdata;
		info.rinfo[2] = (void*)&num;

		info.prim.nmflags = 0;
		info.prim.nmflags = DEF_NMFLAGS_REQ_NAMEQUERY;
		

		res = send_RequestMessageRepeatAnyTimes(&info); // unicast
		ct->tid = info.prim.TranID;
		if(res != LIBNBT_ERR_NOERR){
			if(res == LIBNBT_ERR_INTERRUPTED){
				return res;
			}
			if(res != LIBNBT_ERR_NORESPONSE){
				D_PRINTF(("[libnbt] addname_addgroup_p_node: send_RequestMessageRepeatAnyTimes failed\n"));
				return LIBNBT_ERR_OTHER_ERR;
			}

		}
		if((res == LIBNBT_ERR_NORESPONSE) || (info.retval == NBT_NS_MT_RES_NEG_NAMEQUERY)){
			// no response or NEGATIVE
			// update
			// name can be claimed.
			
			// create
			info.mestype 		= NBT_NS_MT_REQ_REF;
			info.dstnaddr 		= winsnaddr;
			info.ignoreresponse	= NS_RESTYPE_ENODECHA;
			info.prim.nmflags	= DEF_NMFLAGS_REQ_REL;
			info.prim.TTL		= 0;
			
			res = send_RequestMessageRepeatAnyTimes(&info);
			ct->tid = info.prim.TranID;
			if(res != LIBNBT_ERR_NOERR){
				if(res == LIBNBT_ERR_INTERRUPTED){
					return res;
				}
				if(res != LIBNBT_ERR_NORESPONSE){
					D_PRINTF(("[libnbt] addname_addgroup_p_node: send_RequestMessageRepeatAnyTimes failed\n"));
					return LIBNBT_ERR_OTHER_ERR;
				}
			}
			
			if(res == LIBNBT_ERR_NORESPONSE){
				// no response
				// name could not be claimed
				*rcode = (unsigned char)-1;
				return LIBNBT_ERR_NAMEREG_FAILED; // failure
			}
			
			// response
			switch(info.retval){
			case NBT_NS_MT_RES_POS_NAMEREG:
				// add name
				*refresh = rinfo.TTL;
				return LIBNBT_ERR_NOERR; // success
			
			case NBT_NS_MT_RES_NEG_NAMEREG:
				// you lose...
				/* read rcode */
				c = getGroupMacroFromValue(info.prim.Rcode);
				if(c == (unsigned char)-1){
					return LIBNBT_ERR_OTHER_ERR;
				}
				return LIBNBT_ERR_NAMEREG_FAILED; // failure
				
			default:
				D_PRINTF(("[libnbt] addname_addgroup_p_node: send_RequestMessageRepeatAnyTimes no such retval: %d\n",
					info.retval));
				return LIBNBT_ERR_OTHER_ERR;
			}
			
			// never come here
		} // no response
		
		// response
		if(info.retval != NBT_NS_MT_RES_POS_NAMEQUERY){
			D_PRINTF(("[libnbt] addname_addgroup_p_node: info.retval != POSITIVE_NAME_QUERY_RESPONSE\n"));
			return LIBNBT_ERR_OTHER_ERR;
		}
		
		// Remote node still has name
		*rcode = (unsigned char)-1;
		return LIBNBT_ERR_NAMEREG_FAILED; // failure
		
	default:
		D_PRINTF(("[libnbt] addname_addgroup_p_node: send_RequestMessageRepeatAnyTimes no such retval\n"));
		return LIBNBT_ERR_OTHER_ERR;
	}
	
	// never come here
	return LIBNBT_ERR_OTHER_ERR;
	
}


// addname, addgroup for B-node
/*
------------------------------------------------------------------
addname_b_node()

B-nodepOo^֐


߂l:
	:	
		LIBNBT_ERR_NOERR

	s:
		LIBNBT_ERR_INVALID_PARAM
		LIBNBT_ERR_INTERRUPTED
		LIBNBT_ERR_NAMEREG_FAILED... rcodeɒlݒ肳(rcode == -1: Ȃ)
		LIBNBT_ERR_OTHER_ERROR

------------------------------------------------------------------
*/
static int addname_b_node(NBTCTBL* ct, const NBTNamePack* np, unsigned long* refresh, int mestype,
						  unsigned char* rcode, NBTForceSettingInfo* fcinfo)
{

	*refresh = 0;
	return addname_broadcast(ct, np, mestype, rcode, fcinfo);
}



// addname, addgroup for P-node
/*
------------------------------------------------------------------
addname_p_node()

P-nodepOo^֐

߂l:
	:	
		LIBNBT_ERR_NOERR

	s:
		LIBNBT_ERR_INVALID_PARAM
		LIBNBT_ERR_INTERRUPTED
		LIBNBT_ERR_NAMEREG_FAILED... rcodeɒlݒ肳(rcode == -1: Ȃ)
		LIBNBT_ERR_OTHER_ERROR

------------------------------------------------------------------
*/
static int addname_p_node(NBTCTBL* ct, const NBTNamePack* np, unsigned long winsnaddr,
						  unsigned long* refresh, int mestype, unsigned char* rcode,
						  NBTForceSettingInfo* fcinfo)
{
	*refresh = 0;
	return addname_unicast(ct, np, winsnaddr, refresh, mestype, rcode, fcinfo);
}



// M-node: B-node --> P-node
/*
------------------------------------------------------------------
addname_m_node()

M-nodepOo^֐

߂l:
	:	
		LIBNBT_ERR_NOERR

	s:
		LIBNBT_ERR_INVALID_PARAM
		LIBNBT_ERR_INTERRUPTED
		LIBNBT_ERR_NAMEREG_FAILED... rcodeɒlݒ肳(rcode == -1: Ȃ)
		LIBNBT_ERR_OTHER_ERROR

------------------------------------------------------------------
*/
static int addname_m_node(NBTCTBL* ct, const NBTNamePack* np, unsigned long winsnaddr,
						  unsigned long* refresh, int mestype, unsigned char* rcode,
						  NBTForceSettingInfo* fcinfo)
{
	int res;
	
	*refresh = 0;
	/* u[hLXgꍇ̂NBNSɓo^ */
	res = addname_broadcast(ct, np, mestype, rcode, fcinfo);
	if(res == LIBNBT_ERR_NOERR){
		res = addname_unicast(ct, np, winsnaddr, refresh, mestype, rcode, fcinfo);
	}
	return res;
}



// H-node: P-node, if failed then B-node
/*
------------------------------------------------------------------
addname_h_node()

H-nodepOo^֐

߂l:
	:	
		LIBNBT_ERR_NOERR

	s:
		LIBNBT_ERR_INVALID_PARAM
		LIBNBT_ERR_INTERRUPTED
		LIBNBT_ERR_NAMEREG_FAILED... rcodeɒlݒ肳(rcode == -1: Ȃ)
		LIBNBT_ERR_OTHER_ERROR

------------------------------------------------------------------
*/
static int addname_h_node(NBTCTBL* ct, const NBTNamePack* np, unsigned long winsnaddr,
						  unsigned long* refresh, int mestype, unsigned char* rcode,
						  NBTForceSettingInfo* fcinfo)
{
	int res;
	
	*refresh = 0;
	
	res = addname_unicast(ct, np, winsnaddr, refresh, mestype, rcode, fcinfo);
	if(res != LIBNBT_ERR_NOERR){
		if(res == LIBNBT_ERR_INTERRUPTED){	/* f̎͏I */
			return res;
		}

		res = addname_broadcast(ct, np, mestype, rcode, fcinfo);
	}

	return res;
}



/*
------------------------------------------------------------------
nbt_addname()

Oo^sB

߂l:
	:	
		LIBNBT_ERR_NOERR

	s:
		LIBNBT_ERR_INVALID_PARAM
		LIBNBT_ERR_INTERRUPTED
		LIBNBT_ERR_NAMEREG_FAILED... rcodeɒlݒ肳(rcode == -1: Ȃ)
		LIBNBT_ERR_OTHER_ERROR

------------------------------------------------------------------
*/
int nbt_addname(int s, const NBTNamePack* np, unsigned long* refresh, unsigned char* rcode,
				NBTForceSettingInfo* fcinfo)
{
	NBTCTBL* ct;
	int res;
	unsigned long winsnaddr = 0;

	if((np == NULL) || (refresh == NULL) || (rcode == NULL)){
		return LIBNBT_ERR_INVALID_PARAM;
	}

	ct = getCtbl(decListNumber(s));
	if(!ct){
		return LIBNBT_ERR_INVALID_PARAM;
	}
	if(ct->serviceType != NBT_STYPE_NAME){
		return LIBNBT_ERR_INVALID_PARAM;
	}

	/* fBXNv^ɐݒ肳ꂽWINST[oAhXDIɎgp */
	if(ct->winsnaddr != 0){
		winsnaddr = ct->winsnaddr;
	}
	else if(GSetting.winsnaddr != 0){
		winsnaddr = GSetting.winsnaddr;
	}


	// node-typeƂɈقȂ鏈
	switch(GSetting.ont){
	case NBT_ONT_B:
		res = addname_b_node(ct, np, refresh, NBT_NS_MT_REQ_NAMEREG, rcode, fcinfo);
		break;
	case NBT_ONT_P:
		res = addname_p_node(ct, np, winsnaddr, refresh, NBT_NS_MT_REQ_NAMEREG, rcode, fcinfo);
		break;
	case NBT_ONT_M:
		res = addname_m_node(ct, np, winsnaddr, refresh, NBT_NS_MT_REQ_NAMEREG, rcode, fcinfo);
		break;
	case NBT_ONT_H:
		res = addname_h_node(ct, np, winsnaddr, refresh, NBT_NS_MT_REQ_NAMEREG, rcode, fcinfo);
		break;
	default:
		D_PRINTF(("[libnbt] nbt_addname: no such node type. %d\n", GSetting.ont));
		return LIBNBT_ERR_INVALID_PARAM;
	}

	return res;
}



/*
------------------------------------------------------------------
nbt_addname_multihomed()

multihomedOo^s

߂l:
	:	
		LIBNBT_ERR_NOERR

	s:
		LIBNBT_ERR_INVALID_PARAM
		LIBNBT_ERR_INTERRUPTED
		LIBNBT_ERR_NAMEREG_FAILED... rcodeɒlݒ肳(rcode == -1: Ȃ)
		LIBNBT_ERR_OTHER_ERROR

------------------------------------------------------------------
*/
int nbt_addname_multihomed(int s, const NBTNamePack* np, unsigned long* refresh, unsigned char* rcode,
						   NBTForceSettingInfo* fcinfo)
{
	NBTCTBL* ct;
	int res;
	unsigned long winsnaddr = 0;

	if((np == NULL) || (refresh == NULL) || (rcode == NULL)){
		return LIBNBT_ERR_INVALID_PARAM;
	}

	ct = getCtbl(decListNumber(s));
	if(!ct){
		return LIBNBT_ERR_INVALID_PARAM;
	}
	if(ct->serviceType != NBT_STYPE_NAME){
		return LIBNBT_ERR_INVALID_PARAM;
	}

	/* fBXNv^ɐݒ肳ꂽWINST[oAhXDIɎgp */
	if(ct->winsnaddr != 0){
		winsnaddr = ct->winsnaddr;
	}
	else if(GSetting.winsnaddr != 0){
		winsnaddr = GSetting.winsnaddr;
	}


	// node-typeƂɈقȂ鏈
	switch(GSetting.ont){
	case NBT_ONT_B:
		res = addname_b_node(ct, np, refresh, NBT_NS_MT_REQ_MHNAMEREG, rcode, fcinfo);
		break;
	case NBT_ONT_P:
		res = addname_p_node(ct, np, winsnaddr, refresh, NBT_NS_MT_REQ_MHNAMEREG, rcode, fcinfo);
		break;
	case NBT_ONT_M:
		res = addname_m_node(ct, np, winsnaddr, refresh, NBT_NS_MT_REQ_MHNAMEREG, rcode, fcinfo);
		break;
	case NBT_ONT_H:
		res = addname_h_node(ct, np, winsnaddr, refresh, NBT_NS_MT_REQ_MHNAMEREG, rcode, fcinfo);
		break;
	default:
		D_PRINTF(("[libnbt] nbt_addname: no such node type.\n"));
		return LIBNBT_ERR_INVALID_PARAM;
	}

	return res;

}


// low level
/* 1̃x̃t@NV */
/*
------------------------------------------------------------------
nbt_addname_unicast()

jLXgŖOo^s

߂l:
	:	
		LIBNBT_ERR_NOERR

	s:
		LIBNBT_ERR_INVALID_PARAM
		LIBNBT_ERR_INTERRUPTED
		LIBNBT_ERR_NAMEREG_FAILED... rcodeɒlݒ肳(rcode == -1: Ȃ)
		LIBNBT_ERR_OTHER_ERROR

------------------------------------------------------------------
*/
int nbt_addname_unicast(int s, const NBTNamePack* np, unsigned long winsnaddr,
						unsigned long* refresh, unsigned char* rcode, NBTForceSettingInfo* fcinfo)
{
	NBTCTBL* ct;

	if((np == NULL) || (refresh == NULL) || (rcode == NULL)){
		return LIBNBT_ERR_INVALID_PARAM;
	}
	ct = getCtbl(decListNumber(s));
	if(!ct){
		return LIBNBT_ERR_INVALID_PARAM;
	}
	if(ct->serviceType != NBT_STYPE_NAME){
		return LIBNBT_ERR_INVALID_PARAM;
	}

	return addname_unicast(ct, np, winsnaddr, refresh,
		NBT_NS_MT_REQ_NAMEREG, rcode, fcinfo);

}


/*
------------------------------------------------------------------
nbt_addname_multihomed_unicast()

jLXgmultihomedOo^s

߂l:
	:	
		LIBNBT_ERR_NOERR

	s:
		LIBNBT_ERR_INVALID_PARAM
		LIBNBT_ERR_INTERRUPTED
		LIBNBT_ERR_NAMEREG_FAILED... rcodeɒlݒ肳(rcode == -1: Ȃ)
		LIBNBT_ERR_OTHER_ERROR

------------------------------------------------------------------
*/
int nbt_addname_multihomed_unicast(int s, const NBTNamePack* np, unsigned long winsnaddr,
								   unsigned long* refresh, unsigned char* rcode, NBTForceSettingInfo* fcinfo)
{
	NBTCTBL* ct;

	if((np == NULL) || (refresh == NULL) || (rcode == NULL)){
		return LIBNBT_ERR_INVALID_PARAM;
	}

	ct = getCtbl(decListNumber(s));
	if(!ct){
		return LIBNBT_ERR_INVALID_PARAM;
	}
	if(ct->serviceType != NBT_STYPE_NAME){
		return LIBNBT_ERR_INVALID_PARAM;
	}

	return addname_unicast(ct, np, winsnaddr, refresh, NBT_NS_MT_REQ_MHNAMEREG, rcode, fcinfo);
}


/*
------------------------------------------------------------------
nbt_addname_broadcast()

u[hLXgŖOo^s

߂l:
	:
		LIBNBT_ERR_NOERR

	s:
		LIBNBT_ERR_INVALID_PARAM
		LIBNBT_ERR_INTERRUPTED
		LIBNBT_ERR_NAMEREG_FAILED... rcodeɒlݒ肳(rcode == -1: Ȃ)
		LIBNBT_ERR_OTHER_ERROR

------------------------------------------------------------------
*/
int nbt_addname_broadcast(int s, const NBTNamePack* np, unsigned char* rcode,
						  NBTForceSettingInfo* fcinfo)
{
	NBTCTBL* ct;

	if((np == NULL) || (rcode == NULL)){
		return LIBNBT_ERR_INVALID_PARAM;
	}

	ct = getCtbl(decListNumber(s));
	if(!ct){
		return LIBNBT_ERR_INVALID_PARAM;
	}
	if(ct->serviceType != NBT_STYPE_NAME){
		return LIBNBT_ERR_INVALID_PARAM;
	}

	return addname_broadcast(ct, np, NBT_NS_MT_REQ_NAMEREG, rcode, fcinfo);

}

////////////// findname

/*
------------------------------------------------------------------
findname_broadcast()

u[hLXgŖOB

߂l:
	:
		LIBNBT_ERR_NOERR

	s:		
		LIBNBT_ERR_INTERRUPTED
		LIBNBT_ERR_NAME_NOTFOUND
		LIBNBT_ERR_OTHER_ERROR

------------------------------------------------------------------
*/
static int findname_broadcast(NBTCTBL* ct, const NBTNamePack* np, NBTAddrPack* ap, int num, int* findnum,
							  NBTForceSettingInfo* fcinfo)
{
	stRequestMessageInfo info;
	stNameServicePrimitiveRequest rinfo;
	int res;
	int findnum_local;
	
	if(findnum){
		*findnum = 0;
	}
	if(ap && (num > 0)){
		memset(ap, 0, sizeof(NBTAddrPack)*num);
	}

	// create
	memset(&info, 0, sizeof info);
	info.fd 			= ct->fd;
	info.mestype 		= NBT_NS_MT_REQ_NAMEQUERY;
	info.dstnaddr 		= ct->bcnaddr;
	info.dstnport 		= htons(NBT_PORT_NAME);
	info.retry 			= fcinfo? ((fcinfo->bcast_retry_count != -1)?
								fcinfo->bcast_retry_count: BCAST_REQ_RETRY_COUNT): BCAST_REQ_RETRY_COUNT;
	info.timeout		= fcinfo? ((fcinfo->bcast_retry_timeo != ~0)?
								fcinfo->bcast_retry_timeo: BCAST_REQ_RETRY_TIMEOUT): BCAST_REQ_RETRY_TIMEOUT;

	info.waitresponse	= 1;
	info.ignoreresponse	= NS_RESTYPE_REDIRECT | NS_RESTYPE_NEGATIVE;
	info.rinfo[0] 		= (void*)&rinfo;
	if(ap && (num > 0)){
		info.rinfo[1]	= (void*)ap;
		info.rinfo[2]	= (void*)&num;
		info.rinfo[3]	= (void*)&findnum_local;
	}
	else{
		int tnum = 0;
		info.rinfo[1]	= (void*)NULL;
		info.rinfo[2]	= (void*)&tnum;
		info.rinfo[3]	= (void*)&findnum_local;
	}

	info.interruptd		= fcinfo? ((fcinfo->interruptd != -1)? fcinfo->interruptd: -1): -1;
	info.prim.TranID 	= ct->tid;
	info.prim.TTL		= 0;

	nbt_padname(np->name, info.prim.QName, NBT_LENGTH_NETBIOSNAME, ' ', np->suffix);

	strcpy(info.prim.QScopeId, ct->scopeid? ct->scopeid: GSetting.scopeid);	// scopeid

	info.prim.ONT			= GSetting.ont;
	info.prim.G				= 0; // DONT CARE
	info.prim.nmflags		= NS_NSPR_NMFLAGS_B_ON | DEF_NMFLAGS_REQ_NAMEQUERY;
	
	res = send_RequestMessageRepeatAnyTimes(&info);
	ct->tid = info.prim.TranID;
	if(res != LIBNBT_ERR_NOERR){
		D_PRINTF(("[libnbt] findname_broadcast: send_RequestMessageRepeatAnyTimes failed\n"));
		if(res == LIBNBT_ERR_INTERRUPTED){
			return res;
		}
		if(res == LIBNBT_ERR_NORESPONSE){
			return LIBNBT_ERR_NAME_NOTFOUND;
		}

		return LIBNBT_ERR_OTHER_ERR;
	}
	

	switch(info.retval){
	case NBT_NS_MT_RES_POS_NAMEQUERY:
		// Start a timer to detect conflict.
		/* ꖼǑosȂKv͂邩H */
		// readable??
		//D_PRINTF(("[libnbt] findname_broadcast: INFO: allEntries=%d, readEntries=%d, G=%d, ONT=%d, nbaddr=%s\n", 
		//	rinfo.NSPR_nentries, rinfo.NSPR_nentries_read, .G, rdata.ONT, inet_ntoa(*(struct in_addr*)&rdata.nbaddr)));
		if(findnum){
			*findnum = findnum_local;
		}
		return LIBNBT_ERR_NOERR; // success

	default:
		D_PRINTF(("[libnbt] findname_broadcast: send_RequestMessageRepeatAnyTimes no such retval: %d\n", info.retval));
		return LIBNBT_ERR_OTHER_ERR;
	}

	// never come here.
	return LIBNBT_ERR_OTHER_ERR;
}




/*
------------------------------------------------------------------
findname_unicast()

jLXgŖOB

߂l:
	:
		LIBNBT_ERR_NOERR

	s:
		LIBNBT_ERR_INVALID_PARAM
		LIBNBT_ERR_INTERRUPTED
		LIBNBT_ERR_OTHER_ERROR
		LIBNBT_ERR_NAME_NOTFOUND

------------------------------------------------------------------
*/
static int findname_unicast(NBTCTBL* ct, const NBTNamePack* np, unsigned long winsnaddr,
							NBTAddrPack* ap, int num, int* findnum, NBTForceSettingInfo* fcinfo)
{
	stRequestMessageInfo info;
	stNameServicePrimitiveRequest rinfo;
	int res;
	int findnum_local;

	if(winsnaddr == 0){
		return LIBNBT_ERR_INVALID_PARAM;	// failure
	}

	if(findnum){
		*findnum = 0;
	}

	if(ap && (num > 0)){
		memset(ap, 0, sizeof(NBTAddrPack)*num);
	}

	// create
	memset(&info, 0, sizeof info);
	info.fd 			= ct->fd;
	info.mestype 		= NBT_NS_MT_REQ_NAMEQUERY;
	info.dstnaddr 		= winsnaddr;
	info.dstnport 		= htons(NBT_PORT_NAME);
	info.retry 			= fcinfo? ((fcinfo->ucast_retry_count != -1)?
							fcinfo->ucast_retry_count: UCAST_REQ_RETRY_COUNT): UCAST_REQ_RETRY_COUNT;

	info.timeout		= fcinfo? ((fcinfo->ucast_retry_timeo != ~0)?
							fcinfo->ucast_retry_timeo: UCAST_REQ_RETRY_TIMEOUT): UCAST_REQ_RETRY_TIMEOUT;
	info.waitresponse	= 1;
	info.ignoreresponse	= 0;
	info.rinfo[0] 		= (void*)&rinfo;
	if(ap && (num > 0)){
		info.rinfo[1]	= (void*)ap;
		info.rinfo[2]	= (void*)&num;
		info.rinfo[3]	= (void*)&findnum_local;
	}
	else{
		int tnum = 0;
		info.rinfo[1]	= (void*)NULL;
		info.rinfo[2]	= (void*)&tnum;
		info.rinfo[3]	= (void*)&findnum_local;
	}

	info.interruptd		= fcinfo? ((fcinfo->interruptd != -1)? fcinfo->interruptd: -1): -1;
	info.prim.TranID 	= ct->tid;
	info.prim.TTL		= 0;
	nbt_padname(np->name, info.prim.QName, NBT_LENGTH_NETBIOSNAME, ' ', np->suffix);

	strcpy(info.prim.QScopeId, ct->scopeid? ct->scopeid: GSetting.scopeid);	// scopeid

	info.prim.ONT			= GSetting.ont;
	info.prim.G				= 0; // DONT CARE
	info.prim.nmflags		= DEF_NMFLAGS_REQ_NAMEQUERY;

	while(1){

		res = send_RequestMessageRepeatAnyTimes(&info);
		ct->tid = info.prim.TranID;
		if(res != LIBNBT_ERR_NOERR){
			if(res == LIBNBT_ERR_INTERRUPTED){
				return res;
			}
			D_PRINTF(("[libnbt] findname_unicast: send_RequestMessageRepeatAnyTimes failed\n"));
			return LIBNBT_ERR_OTHER_ERR;
		}
		

		switch(info.retval){
		case NBT_NS_MT_RES_POS_NAMEQUERY:
			// D_PRINTF(("[libnbt] findname_unicast: INFO: allEntries=%d, readEntries=%d, G=%d, ONT=%d, nbaddr=%s\n", 
			//  rinfo.NSPR_nentries, rinfo.NSPR_nentries_read, rdata.G, rdata.ONT, inet_ntoa(*(struct in_addr*)&rdata.nbaddr)));
			if(findnum){
				*findnum = findnum_local;
			}

			return LIBNBT_ERR_NOERR; // success

		case NBT_NS_MT_RES_REDNAMEQUERY:
			// NBNS node wants this end node to use some other NBNS node to resolve the query.
			// info.dstnaddr = 
			// continue;
			/* ̏ΉĂ܂ */
			return LIBNBT_ERR_NAME_NOTFOUND; // failure

		case NBT_NS_MT_RES_NEG_NAMEQUERY:
			return LIBNBT_ERR_NAME_NOTFOUND; // failure

		default:
			D_PRINTF(("[libnbt] findname_unicast: send_RequestMessageRepeatAnyTimes no such retval\n"));
			return LIBNBT_ERR_OTHER_ERR;

		}
	} // while

	// never come here.
	return LIBNBT_ERR_OTHER_ERR;
}


// B-node
/*
------------------------------------------------------------------
findname_b_node()

B-nodepOo^[`

߂l:
	:
		LIBNBT_ERR_NOERR

	s:		
		LIBNBT_ERR_INTERRUPTED
		LIBNBT_ERR_NAME_NOTFOUND
		LIBNBT_ERR_OTHER_ERROR

------------------------------------------------------------------
*/
static int findname_b_node(NBTCTBL* ct, const NBTNamePack* np, NBTAddrPack* ap, int num, int* findnum,
						   NBTForceSettingInfo* fcinfo)
{
	return findname_broadcast(ct, np, ap, num, findnum, fcinfo);
}

// P-node
/*
------------------------------------------------------------------
findname_p_node()

P-nodepOo^[`

߂l:
	:
		LIBNBT_ERR_NOERR

	s:
		LIBNBT_ERR_INVALID_PARAM
		LIBNBT_ERR_INTERRUPTED
		LIBNBT_ERR_OTHER_ERROR
		LIBNBT_ERR_NAME_NOTFOUND

------------------------------------------------------------------
*/
static int findname_p_node(NBTCTBL* ct, const NBTNamePack* np, unsigned long winsnaddr,
						   NBTAddrPack* ap, int num, int* findnum,
						   NBTForceSettingInfo* fcinfo)
{
	return findname_unicast(ct, np, winsnaddr, ap, num, findnum, fcinfo);
}

// M-node: B-node --> P-node
/*
------------------------------------------------------------------
findname_m_node()

M-nodepOo^[`

߂l:
	:
		LIBNBT_ERR_NOERR

	s:
		LIBNBT_ERR_INVALID_PARAM
		LIBNBT_ERR_INTERRUPTED
		LIBNBT_ERR_OTHER_ERROR
		LIBNBT_ERR_NAME_NOTFOUND

------------------------------------------------------------------
*/
static int findname_m_node(NBTCTBL* ct, const NBTNamePack* np, unsigned long winsnaddr,
						   NBTAddrPack* ap, int num, int* findnum,
						   NBTForceSettingInfo* fcinfo)
{
	int res;

	/* u[hLXgŌČȂNBNSɌɍs */
	res = findname_broadcast(ct, np, ap, num, findnum, fcinfo);
	if(res != LIBNBT_ERR_NOERR){
		res = findname_unicast(ct, np, winsnaddr, ap, num, findnum, fcinfo);
	}

	return res;
}

// H-node: P-node --> B-node
/*
------------------------------------------------------------------
findname_h_node()

H-nodepOo^[`

߂l:
	:
		LIBNBT_ERR_NOERR

	s:
		LIBNBT_ERR_INVALID_PARAM
		LIBNBT_ERR_INTERRUPTED
		LIBNBT_ERR_OTHER_ERROR
		LIBNBT_ERR_NAME_NOTFOUND

------------------------------------------------------------------
*/
static int findname_h_node(NBTCTBL* ct, const NBTNamePack* np, unsigned long winsnaddr,
						   NBTAddrPack* ap, int num, int* findnum,
						   NBTForceSettingInfo* fcinfo)
{
	int res;

	// NBNSɌɍsȂAu[hLXgɂČ
	res = findname_unicast(ct, np, winsnaddr, ap, num, findnum, fcinfo);
	if(res != LIBNBT_ERR_NOERR){
		res = findname_broadcast(ct, np, ap, num, findnum, fcinfo);
	}

	return res;
}


/*
------------------------------------------------------------------
nbt_findname()

Os

߂l:
	:
		LIBNBT_ERR_NOERR

	s:
		LIBNBT_ERR_INVALID_PARAM
		LIBNBT_ERR_INTERRUPTED
		LIBNBT_ERR_OTHER_ERROR
		LIBNBT_ERR_NAME_NOTFOUND

------------------------------------------------------------------
*/
int nbt_findname(int s, const NBTNamePack* np, NBTAddrPack* ap, int num, int* findnum,
				 NBTForceSettingInfo* fcinfo)
{
	NBTCTBL* ct;
	ct = getCtbl(decListNumber(s));
	if(!ct){
		return LIBNBT_ERR_INVALID_PARAM;
	}
	if(ct->serviceType != NBT_STYPE_NAME){
		return LIBNBT_ERR_INVALID_PARAM;
	}

	return in_findname(ct, np, ap, num, findnum, fcinfo);
}


// parameter check: exist
// ap == NULL or findnum == NULL: ok
/*
------------------------------------------------------------------
in_findname()

Oigpj


߂l:
	:
		LIBNBT_ERR_NOERR

	s:
		LIBNBT_ERR_INVALID_PARAM
		LIBNBT_ERR_INTERRUPTED
		LIBNBT_ERR_OTHER_ERROR
		LIBNBT_ERR_NAME_NOTFOUND

------------------------------------------------------------------
*/
int in_findname(NBTCTBL* ct, const NBTNamePack* np, NBTAddrPack* ap, int num, int* findnum, NBTForceSettingInfo* fcinfo)
{
	int res;
	unsigned long winsnaddr = 0;

	if((ct == NULL) || (np == NULL)){
		return LIBNBT_ERR_INVALID_PARAM;
	}
	if(ct->messageType != NBT_STYPE_NAME){
		return LIBNBT_ERR_INVALID_PARAM;
	}

	/* fBXNv^ɐݒ肳ꂽWINST[oAhXDIɎgp */
	if(ct->winsnaddr != 0){
		winsnaddr = ct->winsnaddr;
	}
	else if(GSetting.winsnaddr != 0){
		winsnaddr = GSetting.winsnaddr;
	}
	
	// node-typeƂɈقȂ鏈
	switch(GSetting.ont){
	case NBT_ONT_B:
		res = findname_b_node(ct, np, ap, num, findnum, fcinfo);
		break;
	case NBT_ONT_P:
		res = findname_p_node(ct, np, winsnaddr, ap, num, findnum, fcinfo);
		break;
	case NBT_ONT_M:
		res = findname_m_node(ct, np, winsnaddr, ap, num, findnum, fcinfo);
		break;
	case NBT_ONT_H:
		res = findname_h_node(ct, np, winsnaddr, ap, num, findnum, fcinfo);
		break;
	default:
		D_PRINTF(("[libnbt] in_findname: no such node type.\n"));
		return LIBNBT_ERR_INVALID_PARAM;
	}

	return res;
}




// delete name

/*
------------------------------------------------------------------
deletename_broadcast()

u[hLXgŖO

߂l:
	:
		LIBNBT_ERR_NOERR

	s:
		LIBNBT_ERR_INTERRUPTED
		LIBNBT_ERR_OTHER_ERROR
		LIBNBT_ERR_DELETE_FAILED

------------------------------------------------------------------
*/
static int deletename_broadcast(NBTCTBL* ct, const NBTNamePack* np, NBTForceSettingInfo* fcinfo)
{
	int res;
	stRequestMessageInfo info;
	stNameServicePrimitiveRequest rinfo;
	
	// create
	memset(&info, 0, sizeof info);
	info.fd 			= ct->fd;
	info.mestype 		= NBT_NS_MT_REQ_REL;
	info.dstnaddr 		= ct->bcnaddr;
	info.dstnport 		= htons(NBT_PORT_NAME);
	info.retry 			= fcinfo? ((fcinfo->bcast_retry_count != -1)?
								fcinfo->bcast_retry_count: BCAST_REQ_RETRY_COUNT): BCAST_REQ_RETRY_COUNT;

	info.timeout		= fcinfo? ((fcinfo->bcast_retry_timeo != ~0)?
								fcinfo->bcast_retry_timeo: BCAST_REQ_RETRY_TIMEOUT): BCAST_REQ_RETRY_TIMEOUT;
	info.waitresponse	= fcinfo? ((fcinfo->flag & NBT_FORCESET_NORESP)? 0: 1): 1;
	info.ignoreresponse	= 0;
	info.rinfo[0] 		= (void*)&rinfo;
	info.interruptd		= fcinfo? ((fcinfo->interruptd != -1)? fcinfo->interruptd: -1): -1;
	info.prim.TranID 	= ct->tid;
	info.prim.TTL		= 0;

	nbt_padname(np->name, info.prim.QName, NBT_LENGTH_NETBIOSNAME, ' ', np->suffix);

	strcpy(info.prim.QScopeId, ct->scopeid? ct->scopeid: GSetting.scopeid);	// scopeid

	info.prim.NbAddress		= (np->naddr == 0)? ct->srcnaddr: np->naddr;
	info.prim.ONT			= GSetting.ont;
	info.prim.G				= getGroupValueFromMacro(np->group);
	info.prim.nmflags		= NS_NSPR_NMFLAGS_B_ON | DEF_NMFLAGS_REQ_REL;

	res = send_RequestMessageRepeatAnyTimes(&info);
	ct->tid = info.prim.TranID;

	if(info.waitresponse == 0){
		return LIBNBT_ERR_NOERR;	// always success
	}

	if(res == LIBNBT_ERR_INTERRUPTED){
		return res;
	}

	if(res == LIBNBT_ERR_NORESPONSE){
		// no response packet expected.
		return LIBNBT_ERR_NOERR;
	}

	if(res == LIBNBT_ERR_NOERR){
		// any response
		// error
		return LIBNBT_ERR_DELETE_FAILED;
	}

	// other err
	return LIBNBT_ERR_OTHER_ERR;
}



/*
------------------------------------------------------------------
deletename_unicast()

jLXgŖO

߂l:
	:
		LIBNBT_ERR_NOERR

	s:
		LIBNBT_ERR_INVALID_PARAM
		LIBNBT_ERR_INTERRUPTED
		LIBNBT_ERR_OTHER_ERROR
		LIBNBT_ERR_DELETE_FAILED

------------------------------------------------------------------
*/
static int deletename_unicast(NBTCTBL* ct, const NBTNamePack* np, unsigned long winsnaddr, NBTForceSettingInfo* fcinfo)
{
	int res;
	stRequestMessageInfo info;
	stNameServicePrimitiveRequest rinfo;

	if(winsnaddr == 0){
		return LIBNBT_ERR_INVALID_PARAM; // failure
	}

	// create
	D_PRINTF(("winsaddr = %s\n", inet_ntoa(*(struct in_addr*)&winsnaddr)));

	memset(&info, 0, sizeof info);
	info.fd 			= ct->fd;
	info.mestype 		= NBT_NS_MT_REQ_REL;
	info.dstnaddr 		= winsnaddr;
	info.dstnport 		= htons(NBT_PORT_NAME);
	info.retry 			= fcinfo? ((fcinfo->ucast_retry_count != -1)?
							fcinfo->ucast_retry_count: UCAST_REQ_RETRY_COUNT): UCAST_REQ_RETRY_COUNT;
	info.timeout		= fcinfo? ((fcinfo->ucast_retry_timeo != ~0)?
							fcinfo->ucast_retry_timeo: UCAST_REQ_RETRY_TIMEOUT): UCAST_REQ_RETRY_TIMEOUT;

	info.waitresponse	= fcinfo? ((fcinfo->flag & NBT_FORCESET_NORESP)? 0: 1): 1;
	info.ignoreresponse	= 0;
	info.rinfo[0] 		= (void*)&rinfo;
	info.interruptd		= fcinfo? ((fcinfo->interruptd != -1)? fcinfo->interruptd: -1): -1;
	info.prim.TranID 	= ct->tid;
	info.prim.TTL		= 0;
	
	nbt_padname(np->name, info.prim.QName, NBT_LENGTH_NETBIOSNAME, ' ', np->suffix);

	strcpy(info.prim.QScopeId, ct->scopeid? ct->scopeid: GSetting.scopeid);	// scopeid

	info.prim.NbAddress		= (np->naddr == 0)? ct->srcnaddr: np->naddr;
	info.prim.ONT			= GSetting.ont;
	info.prim.G				= getGroupValueFromMacro(np->group);
	info.prim.nmflags		= DEF_NMFLAGS_REQ_REL;

	res = send_RequestMessageRepeatAnyTimes(&info);
	ct->tid = info.prim.TranID;
	if(info.waitresponse == 0){
		return LIBNBT_ERR_NOERR;	// success
	}

	if(res != LIBNBT_ERR_NOERR){
		return LIBNBT_ERR_OTHER_ERR; // failure
	}

	// res == LIBNBT_ERR_NOERR
	switch(info.retval){
	case NBT_NS_MT_RES_POS_NAMEREL:
		return LIBNBT_ERR_NOERR; // success
	case NBT_NS_MT_RES_NEG_NAMEREL:
		return LIBNBT_ERR_DELETE_FAILED; // failure

	default:
		D_PRINTF(("[libnbt] deletename_unicast: send_RequestMessageRepeatAnyTimes  no such response type: %d\n", info.retval));
		return LIBNBT_ERR_OTHER_ERR;
	}

	// never come here
	return LIBNBT_ERR_OTHER_ERR;
}


/*
------------------------------------------------------------------
deletename_b_node()

B-nodepO

------------------------------------------------------------------
*/
static int deletename_b_node(NBTCTBL* ct, const NBTNamePack* np, NBTForceSettingInfo* fcinfo)
{

	return deletename_broadcast(ct, np, fcinfo);
}

/*
------------------------------------------------------------------
deletename_p_node()

P-nodepO

------------------------------------------------------------------
*/
static int deletename_p_node(NBTCTBL* ct, const NBTNamePack* np, unsigned long winsnaddr,
							 NBTForceSettingInfo* fcinfo)
{

	return deletename_unicast(ct, np, winsnaddr, fcinfo);
}

/*
------------------------------------------------------------------
deletenamae_m_node()

M-nodepO

------------------------------------------------------------------
*/
static int deletename_m_node(NBTCTBL* ct, const NBTNamePack* np, unsigned long winsnaddr,
							 NBTForceSettingInfo* fcinfo)
{
	int res;

	// ANBNSւ̖OAu[hLXgsȂB
	res = deletename_unicast(ct, np, winsnaddr, fcinfo);
	if(res == LIBNBT_ERR_NOERR){
		res = deletename_broadcast(ct, np, fcinfo);
	}

	return res;
}


/*
------------------------------------------------------------------
deletename_h_node()

H-nodepO

------------------------------------------------------------------
*/
static int deletename_h_node(NBTCTBL* ct, const NBTNamePack* np, unsigned long winsnaddr,
							 NBTForceSettingInfo* fcinfo)
{
	int res;
	
	// ANBNSւ̖OAu[hLXgsȂB
	res = deletename_unicast(ct, np, winsnaddr, fcinfo);
	if(res == LIBNBT_ERR_NOERR){
		res = deletename_broadcast(ct, np, fcinfo);
	}

	return res;
}


/*
------------------------------------------------------------------
nbt_deletename()

Os

߂l:
	:
		LIBNBT_ERR_NOERR

	s:
		LIBNBT_ERR_INVALID_PARAM
		LIBNBT_ERR_INTERRUPTED
		LIBNBT_ERR_OTHER_ERROR
		LIBNBT_ERR_DELETE_FAILED

------------------------------------------------------------------
*/
int nbt_deletename(int s, const NBTNamePack* np, NBTForceSettingInfo* fcinfo)
{
	NBTCTBL* ct;
	int res;
	unsigned long winsnaddr = 0;

	if(np == NULL){
		return LIBNBT_ERR_INVALID_PARAM;
	}
	ct = getCtbl(decListNumber(s));
	if(!ct){
		return LIBNBT_ERR_INVALID_PARAM;
	}
	if(ct->serviceType != NBT_STYPE_NAME){
		return LIBNBT_ERR_INVALID_PARAM;
	}
	
	/* fBXNv^ɐݒ肳ꂽWINST[oAhXDIɎgp */
	if(ct->winsnaddr != 0){
		winsnaddr = ct->winsnaddr;
	}
	else if(GSetting.winsnaddr != 0){
		winsnaddr = GSetting.winsnaddr;
	}

	// node-typeƂɈقȂ鏈
	switch(GSetting.ont){
	case NBT_ONT_B:
		res = deletename_b_node(ct, np, fcinfo);
		break;
	case NBT_ONT_P:
		res = deletename_p_node(ct, np, winsnaddr, fcinfo);
		break;
	case NBT_ONT_M:
		res = deletename_m_node(ct, np, winsnaddr, fcinfo);
		break;
	case NBT_ONT_H:
		res = deletename_h_node(ct, np, winsnaddr, fcinfo);
		break;
	default:
		D_PRINTF(("[libnbt] nbt_deletegroupname_raw: no such node type: %d\n", GSetting.ont));
		return LIBNBT_ERR_INVALID_PARAM;
	}

	return res;
}


/* 1̃x̊֐ */
/*
------------------------------------------------------------------
nbt_deletename_unicast()

jLXgŖOs

߂l:
	:
		LIBNBT_ERR_NOERR

	s:
		LIBNBT_ERR_INVALID_PARAM
		LIBNBT_ERR_INTERRUPTED
		LIBNBT_ERR_OTHER_ERROR
		LIBNBT_ERR_DELETE_FAILED

------------------------------------------------------------------
*/
int nbt_deletename_unicast(int s, const NBTNamePack* np, unsigned long winsnaddr,
						   NBTForceSettingInfo* fcinfo)
{
	NBTCTBL* ct;

	if(np == NULL){
		return LIBNBT_ERR_INVALID_PARAM;
	}
	ct = getCtbl(decListNumber(s));
	if(!ct){
		return LIBNBT_ERR_INVALID_PARAM;
	}
	if(ct->serviceType != NBT_STYPE_NAME){
		return LIBNBT_ERR_INVALID_PARAM;
	}

	return deletename_unicast(ct, np, winsnaddr, fcinfo);
}



/*
------------------------------------------------------------------
nbt_deletename_broadcast()

u[hLXgŖOs

߂l:
	:
		LIBNBT_ERR_NOERR

	s:
		LIBNBT_ERR_INVALID_PARAM
		LIBNBT_ERR_INTERRUPTED
		LIBNBT_ERR_OTHER_ERROR
		LIBNBT_ERR_DELETE_FAILED

------------------------------------------------------------------
*/
int nbt_deletename_broadcast(int s, const NBTNamePack* np, NBTForceSettingInfo* fcinfo)
{
	NBTCTBL* ct;

	if(np == NULL){
		return LIBNBT_ERR_INVALID_PARAM;
	}

	ct = getCtbl(decListNumber(s));
	if(!ct){
		return LIBNBT_ERR_INVALID_PARAM;
	}
	if(ct->serviceType != NBT_STYPE_NAME){
		return LIBNBT_ERR_INVALID_PARAM;
	}

	return deletename_broadcast(ct, np, fcinfo);
}


// update name
/*
------------------------------------------------------------------
nbt_updatename_unicast()

OXVvijLXgj

߂l:
	:
		LIBNBT_ERR_NOERR

	s:
		LIBNBT_ERR_INVALID_PARAM
		LIBNBT_ERR_INTERRUPTED
		LIBNBT_ERR_OTHER_ERROR

------------------------------------------------------------------
*/
int nbt_updatename_unicast(int s, const NBTNamePack* np, unsigned long winsnaddr, NBTForceSettingInfo* fcinfo)
{
	NBTCTBL* ct;
	int res;
	stRequestMessageInfo info;
	stNameServicePrimitiveRequest rinfo;

	if(np == NULL){
		return LIBNBT_ERR_INVALID_PARAM;
	}
	ct = getCtbl(decListNumber(s));
	if(!ct){
		return LIBNBT_ERR_INVALID_PARAM;
	}
	if(ct->serviceType != NBT_STYPE_NAME){
		return LIBNBT_ERR_INVALID_PARAM;
	}

	if(winsnaddr == 0){
		return LIBNBT_ERR_INVALID_PARAM; // failure
	}

	// create
	memset(&info, 0, sizeof info);
	info.fd 			= ct->fd;
	info.mestype 		= NBT_NS_MT_REQ_REF;
	info.dstnaddr 		= winsnaddr;
	info.dstnport 		= htons(NBT_PORT_NAME);
	info.retry 			= fcinfo? ((fcinfo->ucast_retry_count != -1)?
							fcinfo->ucast_retry_count: UCAST_REQ_RETRY_COUNT): UCAST_REQ_RETRY_COUNT;
	info.timeout		= fcinfo? ((fcinfo->ucast_retry_timeo != ~0)?
							fcinfo->ucast_retry_timeo: UCAST_REQ_RETRY_TIMEOUT): UCAST_REQ_RETRY_TIMEOUT;

	info.waitresponse	= fcinfo? ((fcinfo->flag & NBT_FORCESET_NORESP)? 0: 1): 1;
	info.ignoreresponse	= 0;
	info.rinfo[0] 		= (void*)&rinfo;
	info.interruptd		= fcinfo? ((fcinfo->interruptd != -1)? fcinfo->interruptd: -1): -1;
	info.prim.TranID 	= ct->tid;
	info.prim.TTL		= 0;
	
	nbt_padname(np->name, info.prim.QName, NBT_LENGTH_NETBIOSNAME, ' ', np->suffix);

	strcpy(info.prim.QScopeId, ct->scopeid? ct->scopeid: GSetting.scopeid);	// scopeid

	info.prim.NbAddress		= (np->naddr == 0)? ct->srcnaddr: np->naddr;
	info.prim.ONT			= GSetting.ont;
	info.prim.G				= getGroupValueFromMacro(np->group);
	info.prim.nmflags		= DEF_NMFLAGS_REQ_REF;

	res = send_RequestMessageRepeatAnyTimes(&info);
	ct->tid = info.prim.TranID;
	if(info.waitresponse == 0){
		return 1;	// success
	}

	if(res != LIBNBT_ERR_NOERR){
		return LIBNBT_ERR_OTHER_ERR;
	}


	switch(info.retval){
	case NBT_NS_MT_RES_POS_NAMEREG:
		return LIBNBT_ERR_NOERR; // success
	case NBT_NS_MT_RES_NEG_NAMEREG:
		return LIBNBT_ERR_OTHER_ERR; // failure

	default:
		D_PRINTF(("[libnbt] nbt_updatename_unicast: send_RequestMessageRepeatAnyTimes  no such response type: %d\n",
			info.retval));
		return LIBNBT_ERR_OTHER_ERR;
	}

	// never come here
	return LIBNBT_ERR_OTHER_ERR;

}



/*
------------------------------------------------------------------
nbt_getnodestatus

NodeStatus擾


߂l:
	:
		LIBNBT_ERR_NOERR

	s:
		LIBNBT_ERR_INVALID_PARAM
		LIBNBT_ERR_INTERRUPTED
		LIBNBT_ERR_OTHER_ERROR
		LIBNBT_ERR_MALLOC_FAILED
		LIBNBT_ERR_NAME_NOTFOUND

------------------------------------------------------------------
*/
int nbt_getnodestatus(int s, const NBTNamePack* np,
					NBTNodeName* nodename, int nodename_num, int* get_nodename_num,
					NBTNodeStatistics* statistics,
					NBTForceSettingInfo* fcinfo)
{
	NBTCTBL* ct;
	int res;
	int	i;
	
	stRequestMessageInfo			info;
	stNameServicePrimitiveRequest	rinfo;
	stNameServiceNodeNameEntry*		entry;
	stNameServiceStatistics			stat;
	
	NBTAddrPack		ap;
	int				findnum;
	int				num;
	unsigned long	naddr = 0;
	
	// param check
	
	if(np == NULL){
		return LIBNBT_ERR_INVALID_PARAM;
	}
	
	ct = getCtbl(decListNumber(s));
	if(!ct){
		return LIBNBT_ERR_INVALID_PARAM;
	}
	
	if(ct->serviceType != NBT_STYPE_NAME){
		return LIBNBT_ERR_INVALID_PARAM;
	}
	
	if((np->name[0] == '\0') && (np->naddr == 0)){
		return LIBNBT_ERR_INVALID_PARAM;
	}
	
	if(np->naddr == 0){
		/* IPAhXw肳ĂȂƂfindnamegpIPAhX擾 */
		res = in_findname(ct, np, &ap, 1, &findnum, fcinfo);
		if(res != LIBNBT_ERR_NOERR){
			return res;
		}
		naddr = ap.naddr;
	}
	else{
		naddr = np->naddr;
	}
	
	// create
	memset(&info, 0, sizeof info);
	if(nodename){
		entry = (stNameServiceNodeNameEntry*)malloc(sizeof(stNameServiceNodeNameEntry) * nodename_num);
		if(entry == NULL){
			return LIBNBT_ERR_MALLOC_FAILED;
		}
	}
	else{
		entry = NULL;
	}
	
	info.fd 			= ct->fd;
	info.mestype 		= NBT_NS_MT_REQ_NODESTATUS;
	info.dstnaddr		= naddr;
	info.dstnport		= htons(NBT_PORT_NAME);
	info.retry 			= fcinfo? ((fcinfo->ucast_retry_count != -1)?
							fcinfo->ucast_retry_count: UCAST_REQ_RETRY_COUNT): UCAST_REQ_RETRY_COUNT;
	info.timeout		= fcinfo? ((fcinfo->ucast_retry_timeo != ~0)?
							fcinfo->ucast_retry_timeo: UCAST_REQ_RETRY_TIMEOUT): UCAST_REQ_RETRY_TIMEOUT;
	info.waitresponse	= 1;	/* ɑ肩̉҂ */
	info.ignoreresponse = 0;	/* 鉞͂Ȃ */
	
	num = nodename? nodename_num: 0;
	
	info.rinfo[0]		= (void*)&rinfo;
	info.rinfo[1]		= (void*)entry;
	info.rinfo[2]		= (void*)&num;
	info.rinfo[3]		= (void*)&findnum;
	info.rinfo[4]		= (void*)&stat;
	info.interruptd		= fcinfo? ((fcinfo->interruptd != -1)? fcinfo->interruptd: -1): -1;
	info.prim.TranID	= ct->tid;
	info.prim.TTL		= 0;
	
	/* Ow肳ĂȂƂ"*"gpBw肳ĂƂ͎w肳ꂽOgp */
	if(np->name[0] == '\0'){
		nbt_padname(BROADCAST_NAME, info.prim.QName, NBT_LENGTH_NETBIOSNAME, 0, 0x00);
	}
	else{
		nbt_padname(np->name, info.prim.QName, NBT_LENGTH_NETBIOSNAME, ' ', np->suffix);
	}
	
	strcpy(info.prim.QScopeId, ct->scopeid? ct->scopeid: GSetting.scopeid);	// scopeid
	
	info.prim.nmflags = 0;
	
	res = send_RequestMessageRepeatAnyTimes(&info);
	ct->tid = info.prim.TranID;
	if(res != LIBNBT_ERR_NOERR){
		if(res == LIBNBT_ERR_INTERRUPTED){
			free(entry);
			return res;
		}
		D_PRINTF(("[libnbt] nbt_getnodestatus: send_RequestMessageRepeatAnyTimes failed.\n"));
		free(entry);
		return LIBNBT_ERR_OTHER_ERR;
	}
	
	switch(info.retval){
	case NBT_NS_MT_RES_NODESTATUS:
		*get_nodename_num = findnum;
		
		for(i=0; i<findnum; i++){
			if(i >= num){
				break;
			}
			getNBTNodeNameFromEntry(&nodename[i], &entry[i]);
		}
		getNBTNodeStatisticsFromServiceStatistics(statistics, &stat);
		free(entry);
		return LIBNBT_ERR_NOERR;
		
	default:
		D_PRINTF(("[libnbt] nbt_getnodestatus: send_RequestMessageRepeatAnyTimes no such retval %d.\n", info.retval));
		
		free(entry);
		return LIBNBT_ERR_OTHER_ERR;
	}
	
	// never come here
	return LIBNBT_ERR_OTHER_ERR;
}
