/* -*- tab-width: 4; -*- */
/*------------------------------------------------------------------*
 * Filename:		libnbt_ss.c
 *------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

#ifdef WIN32
#include <WinSock.h>
#include <errno.h>

#else // WIN32
#include <unistd.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/time.h>

#endif // WIN32

#include "libnbt.h"
#include "libnbt_ss.h"
#include "libnbt_private.h"
#include "libnbt_glob.h"
#include "libnbt_mes.h"
#include "libnbt_util.h"
#include "libnbt_util_private.h"
#include "libnbt_ns_in.h"
#include "libnbt_err.h"
#include "libnbt_macros.h"
#include "libnbt_debug.h"

#ifdef WIN32
#define close closesocket

#endif // WIN32

#ifndef WIN32
#ifdef FIXEDLINK
extern int *__errno();
#endif
#endif

// private function def
static unsigned char getMessageTypeToValue(int messageType);
static int getMessageValueToType(unsigned char messageValue);
static int readHeader_ss(NBTCTBL* ct);
static int sendHeader_ss(NBTCTBL* ct, unsigned char* buf, unsigned long len, int mesType);
static int inner_accept(NBTCTBL* ct);
static int inner_send(int s, const void* msg, int len, unsigned int flags, int mesType);

// private
/*
------------------------------------------------------------------
getMessageTypeToValue()

MessageType ۂ̃bZ[Wl֕ϊ

߂l:
	:	ۂ̃bZ[Wl
	s:	-1 (UCHAR_MAX)
------------------------------------------------------------------
*/
static unsigned char getMessageTypeToValue(int messageType)
{
	unsigned char val = -1;
	
	switch(messageType){
	
	// NBT Session Service Type
	case NBT_MTYPE_SES_MES: 		val = VAL_SESSION_MESSAGE; break;
	case NBT_MTYPE_SES_REQ: 		val = VAL_SESSION_REQUEST; break;
	case NBT_MTYPE_SES_RES_ACK: 	val = VAL_SESSION_RESPONSE_ACK; break;
	case NBT_MTYPE_SES_RES_NACK: 	val = VAL_SESSION_RESPONSE_NACK; break;
	case NBT_MTYPE_SES_RES_CHNG: 	val = VAL_SESSION_RESPONSE_CHNG; break;
	case NBT_MTYPE_SES_KPALIV: 		val = VAL_SESSION_KEEPALIVE; break;

	default:
		D_PRINTF(("getMessageTypeToValue: unsupport messageType type = %d\n", messageType));
		break;
	}
	
	return val;
}


/*
------------------------------------------------------------------
getMessageValueToType()

ۂ̃bZ[WlAbZ[W^Cv֕ϊ

:
IN	messageValue:	bZ[Wl

߂l:
	:	bZ[W^Cv
	s:	-1
------------------------------------------------------------------
*/
static int getMessageValueToType(unsigned char messageValue)
{
	int type = -1;
	
	switch(messageValue){
		
	// NBT Session Service Type
	case VAL_SESSION_MESSAGE: 		type = NBT_MTYPE_SES_MES; break;
	case VAL_SESSION_REQUEST: 		type = NBT_MTYPE_SES_REQ; break;
	case VAL_SESSION_RESPONSE_ACK: 	type = NBT_MTYPE_SES_RES_ACK; break;
	case VAL_SESSION_RESPONSE_NACK: type = NBT_MTYPE_SES_RES_NACK; break;
	case VAL_SESSION_RESPONSE_CHNG: type = NBT_MTYPE_SES_RES_CHNG; break;
	case VAL_SESSION_KEEPALIVE: 	type = NBT_MTYPE_SES_KPALIV; break;

	default:
		D_PRINTF(("getMessageValueToType: unsupport messageValue value = 0x%02X\n", messageValue));
		break;
	}
	
	return type;
}


/*
------------------------------------------------------------------
readHeader_ss()

NBTZbVwb_ǂݍ݁Awb_NBTCTBL\̓Ɋi[

:
IN/OUT	ct:	NBTCTBL\̃AhX

߂l:

:
	LIBNBT_ERR_NOERR			(0)

s:
	LIBNBT_ERR_INVALID_PARAM	(-110)
	
	LIBNBT_ERR_RECV_ERR			(-190)
	LIBNBT_ERR_RECVDATA_INVALID	(-501)
	
	LIBNBT_ERR_SELECT_ERR		(-150)
	LIBNBT_ERR_SELECT_TIMEOUT	(-151)

------------------------------------------------------------------
*/
static int readHeader_ss(NBTCTBL* ct)
{
	int res;
	unsigned char buf[4];
	int islisten;
	
	
	/* state check */
	if(	!(	(ct->state == STATE_ACCEPTED_TCP)	||
			(ct->state == STATE_CONNECTED)		||
			(ct->state == STATE_CONNECTED_TCP))){
		
		/* param error */
		return LIBNBT_ERR_INVALID_PARAM;
	}
	
	islisten =
		(ct->state == STATE_LISTENED) | (ct->state == STATE_ACCEPTED_TCP);
	
	res = nbt_readable_checkalive(islisten? ct->cfd: ct->fd, 0);
	if(res < 0){
		return LIBNBT_ERR_SELECT_ERR;
	}
	if(res == 0){
		return LIBNBT_ERR_SELECT_TIMEOUT;
	}
	if(res == 2){
		/* disconnect by remote host */
		ct->messageType = NBT_MTYPE_SES_DISCON;
		ct->messageTotalLen = 0;
		ct->readLen = 0;
		ct->isReadHeader = 1;
		ct->isPacketLast = 1;
		
		return LIBNBT_ERR_NOERR;
	}
	
	res = recv(islisten? ct->cfd: ct->fd, buf, 4, 0);
	if(res < 0){
		return LIBNBT_ERR_RECV_ERR;
	}
	if(res < 4){
		/* f[^s */
		return LIBNBT_ERR_RECVDATA_INVALID;
	}
	
	ct->messageType = getMessageValueToType(buf[0]);
	ct->messageTotalLen = ((buf[1] & 1) << 9) | (buf[2] << 8) | buf[3];
	ct->readLen = 0;
	ct->isReadHeader = 1;
	ct->isPacketLast = ct->messageTotalLen? 0: 1;
	
	if(ct->messageType == -1){
		/* f[^s */
		return LIBNBT_ERR_RECVDATA_INVALID;
	}
	
	return LIBNBT_ERR_NOERR;
}


/*
------------------------------------------------------------------
sendHeader_ss()

NBTZbVwb_obt@ɃZbg

:
IN	ct:			NBTCTBL\
OUT	buf:		wb_Zbgobt@iŒ4oCgKv)
IN	len:		Mf[^̒(byte)
IN	mesType:	bZ[W^Cv

߂l:
:
LIBNBT_ERR_NOERR			(0)

s:
LIBNBT_ERR_INVALID_PARAM	(-110)

------------------------------------------------------------------
*/
static int sendHeader_ss(NBTCTBL* ct, unsigned char* buf, unsigned long len, int mesType)
{

	/* state check */
	if(	!(	(ct->state == STATE_ACCEPTED_TCP)	||
			(ct->state == STATE_CONNECTED)		||
			(ct->state == STATE_CONNECTED_TCP))	){
		
		/* param error */	
		return LIBNBT_ERR_INVALID_PARAM;
	}
	
	if((len >> 17) != 0){
		
		return LIBNBT_ERR_INVALID_PARAM;
	}
	
	// set NBT header
	buf[0] = getMessageTypeToValue(mesType);
	buf[1] = (unsigned char)((len&0x10000)>>16);
	buf[2] = (unsigned char)((len&0xFF00)>>8);
	buf[3] = (unsigned char)(len&0xFF);
	
	if(buf[0] == (unsigned char)-1){
		return LIBNBT_ERR_INVALID_PARAM;
	}
	
	return LIBNBT_ERR_NOERR;
}



/* 
------------------------------------------------------------------
inner_accept()

TCPxł̃RlNVmB
accept֐ĂсAڑ̃\PbgfBXNv^NBTCTBL̃ocfdɑ

NBTCTBL\̂̃ô悤ɏ
isPacketLast = 1	// Mobt@̐擪bZ[WƃbZ[W̋؂łB
isReadHeader = 0	// wb_񖢓ǂݍ

:
IN/OUT	ct:	NBTCTBL\̂̃AhX

߂l:
:
VK쐬ꂽ\PbgfBXNv^

s:
	LIBNBT_ERR_INVALID_PARAM	(-110)
	LIBNBT_ERR_ACCEPT_ERR		(-120)
	LIBNBT_ERR_DISCONNECTED		(-600)
------------------------------------------------------------------
*/
static int inner_accept(NBTCTBL* ct)
{
	int res = -1;
	int addrlen;
	struct sockaddr_in	cliaddr;
	
	/* param check */
	if(ct == NULL){
		return LIBNBT_ERR_INVALID_PARAM;
	}
	
	/* state check */
	if(	!(	(ct->state == STATE_LISTENED)		||
			(ct->state == STATE_ACCEPTED_TCP)	)){
		
		return LIBNBT_ERR_INVALID_PARAM;
	}
				
	addrlen = (int)sizeof(struct sockaddr_in);
	memset(&cliaddr, 0, sizeof(struct sockaddr_in));
	res = accept(ct->fd, (struct sockaddr* )&cliaddr, &addrlen);
	if(res < 0){
		return LIBNBT_ERR_ACCEPT_ERR;
	}
	
	/* NetBSD1.5addrlen0ꍇ肩RSTĂ */
	if(addrlen == 0){
		close(res);
		return LIBNBT_ERR_DISCONNECTED;
	}
	
	/* ȑOaccept̂nbt_acceptĂȂꍇ */
	if(ct->state == STATE_ACCEPTED_TCP){
		close(ct->cfd);	
	}
	ct->cfd = res;
			
	/* LISTENED --> ACCEPTED_TCPɏԑJ */
	ct->state = STATE_ACCEPTED_TCP;
	
	/* bZ[W̎n܂ɂƂӖB */
	ct->isPacketLast = 1;
	ct->isReadHeader = 0;
	
	// connected tcp layer
	
	ct->cnaddr = *(unsigned long* )&cliaddr.sin_addr;
	
	return ct->cfd;
	
}




#define INNERSEND_BUFSIZE	1024

/*
------------------------------------------------------------------
inner_send()

NBTZbVT[rXpbZ[WM֐

{֐ł́Aڑ肪RlNVؒfĂ邩ǂm邱Ƃ͂łȂ

:
IN	s:			NBTfBXNv^
IN	msg:		MbZ[Wobt@
IN	flags:		send֐Ɏw肷tO
IN	mesType:	bZ[W^Cv

߂l:

:
MbZ[W̃oCg

s:
LIBNBT_ERR_INVALID_PARAM	(-110)
LIBNBT_ERR_SELECT_TIMEOUT	(-151)
LIBNBT_ERR_SELECT_ERR		(-150)
LIBNBT_ERR_SENDDATA_INVALID	(-500)
LIBNBT_ERR_SEND_ERR			(-170)

------------------------------------------------------------------
*/
static int inner_send(int s, const void* msg, int len, unsigned int flags, int mesType)
{
	NBTCTBL* ct;
	int res = -1;
	unsigned char buff[INNERSEND_BUFSIZE];
	int osize;
	int wasize;
	int remain;
	int islisten;
	
	unsigned char* p;
	
	
	ct = getCtbl(decListNumber(s));
	if(!ct){
		return LIBNBT_ERR_INVALID_PARAM;
	}
	
	memset(buff, 0, sizeof buff);
	
	/* set header */
	res = sendHeader_ss(ct, buff, len, mesType);
	if(res != LIBNBT_ERR_NOERR){
		return res;
	}
	p = buff + 4;
	wasize = INNERSEND_BUFSIZE-4;
	remain = len;
	
	islisten = (ct->state == STATE_ACCEPTED_TCP);
	
	do{
		res = nbt_writable(islisten? ct->cfd: ct->fd, NBT_WAIT_SEC, -1);
		if(res == 0){
			/* CxgȂ */
			return LIBNBT_ERR_SELECT_TIMEOUT;
		}
		if(res < 0){
			/* select error */
			return LIBNBT_ERR_SELECT_ERR;
		}
		osize = MIN(wasize, remain);	
		memcpy(p, (const u_char*)msg + (len - remain), osize);
		
		res = send(islisten? ct->cfd: ct->fd, buff, osize + (p - buff), flags);
		if(res < 0){
			return LIBNBT_ERR_SEND_ERR;
		}
		if(res != (osize + (p - buff))){
			return LIBNBT_ERR_SENDDATA_INVALID;
		}
		
		// recalc
		p = buff;
		wasize = INNERSEND_BUFSIZE;
		remain -= osize;
		
	} while(remain > 0);
	
	return len;
}



/* J֐S */

/*
------------------------------------------------------------------
nbt_listen()

listen֐R[A\Pbg͐ڑv󂯓鏀

:
	IN s:		nbt_open֐Ŏ擾NBTfBXNv^
	IN backlog:	ɊmłZbV̍ő吔
	
߂l:

:
LIBNBT_ERR_NOERR			(0)

s:
LIBNBT_ERR_INVALID_PARAM	(-110)
LIBNBT_ERR_LISTEN_ERR		(-140)

------------------------------------------------------------------
*/
int nbt_listen(int s, int backlog)
{	
	// listen
	NBTCTBL* ct;
	int res;
	
	ct = getCtbl(decListNumber(s));
	if(ct == NULL){
		return LIBNBT_ERR_INVALID_PARAM;
	}
		
	res = listen(ct->fd, backlog);
	if(res == 0){
		/* state change -> LISTENED */
		ct->state = STATE_LISTENED;
		return LIBNBT_ERR_NOERR;
	}
	
	return LIBNBT_ERR_LISTEN_ERR;
}





/*
------------------------------------------------------------------
nbt_accept()

\Pbgł̐ڑ󂯓AZbVmvNCAg著M̂҂

:
IN	s:		nbt_open֐ɂ擾NBTfBXNv^
OUT	rmt:	ڑzXgNetBIOSEIPAhXݒ肳
IN	proc:	TCPRlNVmɃR[nh̃|C^w肷
	
߂l:
:
	NCAgzXgƃZbVmNBTfBXNv^
	
s:
	LIBNBT_ERR_INVALID_PARAM	(-110)
	LIBNBT_ERR_ACCEPT_ERR		(-200)
	LIBNBT_ERR_NBTFD_MAX		(-120)
	
	LIBNBT_ERR_DISCONNECTED		(-600)
	
in nbt_recv()...
	LIBNBT_ERR_SELECT_ERR		(-150)
	LIBNBT_ERR_SELECT_TIMEOUT	(-151)

	LIBNBT_ERR_RECV_ERR			(-190)
	LIBNBT_ERR_SEND_ERR			(-170)
	
	LIBNBT_ERR_RECVDATA_INVALID	(-501)
	LIBNBT_ERR_SENDDATA_INVALID	(-500)
	
	
	LIBNBT_ERR_RES_NEGATIVE		(-1100)
	
	
ԑJ:
STATE_ACCEPTED_TCP --> STATE_LISTENED
STATE_LISTENED --> STATE_ACCEPTED_TCP --> STATE_LISTENED

VKfBXNv^쐬
STATE_CONNECTED

------------------------------------------------------------------
*/
int nbt_accept(int s, NBTNamePack* rmt, AcceptProc proc)
{
	NBTCTBL* ct;
	NBTCTBL* ctw; // for client fd.
	int res = -1;
	int index;
	int dmy;
	
	u_char buf[512];

	NBTNamePack called;
	NBTNamePack calling;
	
	int readsize;


	if(rmt == NULL){
		return LIBNBT_ERR_INVALID_PARAM;
	}
	
	ct = getCtbl(decListNumber(s));
	if(!ct){
		return LIBNBT_ERR_INVALID_PARAM;
	}
	/* Name Service(TCP) and Session Service. */
	if(!(ct->serviceType == NBT_STYPE_SESSION)){
		return LIBNBT_ERR_INVALID_PARAM;
	}
	
	/* state check */
	if(	!(	(ct->state == STATE_ACCEPTED_TCP)	||
			(ct->state == STATE_LISTENED)	)	){
		
		return LIBNBT_ERR_INVALID_PARAM;
	}
	
	/* listen̏ԂȂ̂ŁAtcpxaccepts */
	if(ct->state == STATE_LISTENED){
		res = inner_accept(ct);
		if(res < 0){
			return res;
		}
	}

	ctw = getUnusedCtbl(&index);
	if(ctw == NULL){
		/* ԂLISTENEDɖ߂ */
		ct->state = STATE_LISTENED;
		
		close(ct->cfd);
		ct->cfd = -1;
		
		/* NBTfBXNv^쐬s\ */
		return LIBNBT_ERR_NBTFD_MAX;
	}
	
	/* VNBTCTBL̈ɏRs[ */
	memcpy(ctw, ct, sizeof(NBTCTBL));
	ctw->fd = ct->cfd;
	
	/* V\Pbg̏ԂCONNECTEDɐݒ肷 */
	ctw->state = STATE_CONNECTED;
	ctw->isUsed = 1;
	
	/* ܂ł̃\Pbg̏ԂLISTENEDɖ߂ */
	ct->state = STATE_LISTENED;
	ct->cfd = -1;
	
	memset(buf, 0, sizeof buf);

	dmy = encListNumber(index);
	
	res = nbt_recv(dmy, buf, sizeof buf, 0, NBT_MSG_ONEMES);
	if(res <= 0){
		close(ctw->fd); initCtbl(ctw);
		return res;
	}
	/* Œf[^ƂāANBT_LENGTH_NETBIOSNAME_2ENC*2̒Ȃ΂ȂȂ */
	if(res < (NBT_LENGTH_NETBIOSNAME_2ENC*2)){
		/* sf[^ */
		close(ctw->fd); initCtbl(ctw);
		
		return LIBNBT_ERR_RECVDATA_INVALID;
	}
	
	memset(&called, 0, sizeof called);
	memset(&calling, 0, sizeof calling);
	
	/* called name()f[^ǂݎ local */
	res = nbt_dec2nd(called.name, NULL, 0, buf, buf, &readsize, NULL, NULL);
	if(res == -1){
		/* ̑̃G[ڑɒm点 */
		nbt_send_nack(dmy, NBT_ERROR_SES_UNSPECIFIED);
		/* G[ǂɊ֌WڑؒfĂ܂ */
		close(ctw->fd);
		initCtbl(ctw);
		
		return LIBNBT_ERR_RECVDATA_INVALID;
	}
	
	/* calling name(M)f[^ǂݎ rmt */
	res = nbt_dec2nd(calling.name, NULL, 0, buf+readsize, buf, &readsize, NULL, NULL);
	if(res == -1){
		/* ̑̃G[ڑɒm点 */
		nbt_send_nack(dmy, NBT_ERROR_SES_UNSPECIFIED);
		/* G[ǂɊ֌WڑؒfĂ܂ */
		close(ctw->fd);
		initCtbl(ctw);
		
		return LIBNBT_ERR_RECVDATA_INVALID;
	}
	
	nbt_devidename(called.name, &called.suffix, called.name);
	nbt_devidename(calling.name, &calling.suffix, calling.name);

	/* S~f[^ꍇ̓bZ[WI[܂œǂݔ΂ */
	res = 1;
	while(res > 0){
		res = nbt_recv(dmy, buf, sizeof buf, 0, NBT_MSG_ONEMES);
	}
	if(res != LIBNBT_ERR_EOM){
		/* ̑̃G[ڑɒm点 */
		nbt_send_nack(dmy, NBT_ERROR_SES_UNSPECIFIED);
		/* G[ǂɊ֌WڑؒfĂ܂ */
		close(ctw->fd);
		initCtbl(ctw);
		
		return res;
	}
	
	res = NBT_ERROR_SES_SUCCESS;
	
	called.naddr = ct->srcnaddr;
	calling.naddr = ctw->cnaddr;
	
	/* nhĂяo */
	if(proc){		
		res = (*proc)(&called, &calling);
	}

	if(res != NBT_ERROR_SES_SUCCESS){
		nbt_send_nack(dmy, res);
		/* G[ǂɊ֌WڑؒfĂ܂ */
		close(ctw->fd);
		initCtbl(ctw);
		
		return LIBNBT_ERR_RES_NEGATIVE;
	}

	/* rmtɑRs[ */
	if(rmt){
		memcpy(rmt, &calling, sizeof calling);
	}
	
	/* m艞sI */
	if((res = nbt_send_ack(dmy)) < 0){
		close(ctw->fd);
		initCtbl(ctw);
		
		return res;
	}
	
	return dmy;
}




/*
------------------------------------------------------------------
nbt_connect()

\Pbg̐ڑAZbV̊msȂ

:
IN	s:		NBTfBXNv^
IN	rmt:	[gzXg̏
IN	local:	[JzXg̏
IN	timeout_msec:	RlNV^CAEgB~b

߂l:
- success
LIBNBT_ERR_NOERR			(0)

- error
LIBNBT_ERR_INVALID_PARAM	(-110)
LIBNBT_ERR_OTHER_ERR		(-1)
LIBNBT_ERR_CONNECT_ERR		(-130)
LIBNBT_ERR_CONNECT_TIMEOUT	(-131)
LIBNBT_ERR_RES_NEGATIVE		(-1100)
LIBNBT_ERR_SELECT_TIMEOUT	(-151)
LIBNBT_ERR_SELECT_ERR		(-150)
LIBNBT_ERR_SENDDATA_INVALID	(-500)
LIBNBT_ERR_SEND_ERR			(-170)
LIBNBT_ERR_ACCEPT_ERR		(-200)
LIBNBT_ERR_RECV_ERR			(-190)
LIBNBT_ERR_RECVDATA_INVALID	(-501)

LIBNBT_ERR_DISCONNECTED		(-600)


ԑJ:
STATE_OPENED --> STATE_CONNECTED_TCP --> STATE_CONNECTED
STATE_CONNECTED_TCP --> STATE_OPENED --> STATE_CONNECTED_TCP --> STATE_CONNECTED
STATE_CONNECTED --> STATE_OPENED --> STATE_CONNECTED_TCP --> STATE_CONNECTED

------------------------------------------------------------------
*/
int nbt_connect(int s, const NBTNamePack* rmt, const NBTNamePack* local,
				unsigned long timeout_msec)
{
	NBTCTBL* ct;
	int res;
	int offset;
	u_long dstnaddr;

	struct sockaddr_in inaddr;
	
	u_char buff[512];
	u_char temp[NBT_LENGTH_NETBIOSNAME];
#ifndef WIN32
	int flags;
#endif
	
	/* parameter check */
	if((rmt == NULL) || (local == NULL)){
		return LIBNBT_ERR_INVALID_PARAM;
	}

	ct = getCtbl(decListNumber(s));
	if(!ct){
		return LIBNBT_ERR_INVALID_PARAM;
	}
	
	/* Session Service Only */
	if(ct->serviceType != NBT_STYPE_SESSION){
		return LIBNBT_ERR_INVALID_PARAM;
	}
	
	/* state check */
	if(	!(	(ct->state == STATE_OPENED)			||
			(ct->state == STATE_CONNECTED_TCP)	||
			(ct->state == STATE_CONNECTED)		||
			(ct->state == STATE_DISCONNECTED)	)){
		
		return LIBNBT_ERR_INVALID_PARAM;
	}


#if 0 /* ̏T|[g */
	/* ڑIPAhX͓͂Ă邩H */
	if(rmt->naddr == 0){
		// Os
		int ns_client;
		NBTAddrPack ap;
		int findnum;
		
		ns_client = nbt_open(NBT_STYPE_NAME, ct->srcnaddr, htons(NBT_PORT_ANY));
		if(ns_client < 0){
			D_PRINTF(("[libnbt] nbt_connect; nbt_open failed\n"));
			return -1;
		}
		res = nbt_findname(ns_client, rmt, &ap, 1, &findnum);
		if(res <= 0){
			D_PRINTF(("[libnbt] nbt_connect: in_findname failed\n"));
			nbt_close(ns_client);
			return -1;
		}
		nbt_close(ns_client);

		if(findnum != 1){
			D_PRINTF(("[libnbt] %d name found\n", findnum));
		}
		
		// u[hLXgAhX͐ڑs
		if(ap.naddr == htonl(0xFFFFFFFFUL)){
			return -1;
		}

		dstnaddr = ap.naddr;
	}
	else{
		dstnaddr = rmt->naddr;
	}
#endif
	dstnaddr = rmt->naddr;
	
	/* 	CONNECTED邢CONNECTED_TCP, DISCONNECTED̏ꍇ́A
		݂̐ڑVbg_EĂVɕʃzXgɐڑ */
	if(	(ct->state == STATE_CONNECTED) 		||
		(ct->state == STATE_CONNECTED_TCP)	||
		(ct->state == STATE_DISCONNECTED)	){
		
		shutdown(ct->fd, 2);
		ct->state = STATE_OPENED;
	}
	
	memset(&inaddr, 0, sizeof(struct sockaddr_in));
#ifndef WIN32
	inaddr.sin_len = sizeof(struct sockaddr_in);
#endif // WIN32

	inaddr.sin_family = AF_INET;
	inaddr.sin_port = htons(NBT_PORT_SESSION);
	inaddr.sin_addr = *(struct in_addr* )&dstnaddr;

	/* socketubN[hɂB */
#ifdef WIN32
	{
		unsigned long on = 1;
		ioctlsocket(ct->fd, FIONBIO, &on);
	}

#else
	flags = fcntl(ct->fd, F_GETFL, 0);
	if(flags != -1){
		res = fcntl(ct->fd, F_SETFL, flags | O_NONBLOCK);
		if(res == -1){
			D_PRINTF(("[libnbt] fcntl failed.\n"));
			return LIBNBT_ERR_OTHER_ERR;
		}
	}
	else{
		D_PRINTF(("[libnbt] fcntl failed.\n"));
		return LIBNBT_ERR_OTHER_ERR;
	}
#endif

	// tcp connection
	res = connect(ct->fd, (struct sockaddr* )&inaddr, sizeof inaddr);
	if(res < 0){
#ifdef WIN32
		if(res != WSAEINPROGRESS){
#else
#ifdef FIXEDLINK
		if(*__errno() != EINPROGRESS){
#else // !FIXEDLINK
		if(errno != EINPROGRESS){
#endif // FIXEDLINK

#endif
			/* socketubN[hɖ߂ */
#ifdef WIN32
			{
				unsigned long off = 0;
				ioctlsocket(ct->fd, FIONBIO, &off);
			}
#else
			fcntl(ct->fd, F_SETFL, flags);
#endif
				
			return LIBNBT_ERR_CONNECT_ERR;
		}
	}
	
	/* ݉\ƂȂƂconnectionmƂ */
	res = nbt_writable(ct->fd, timeout_msec, -1);
	
	/* \PbgubN[hɖ߂ */
#ifdef WIN32
	{
		unsigned long off = 0;
		ioctlsocket(ct->fd, FIONBIO, &off);
	}
#else
	fcntl(ct->fd, F_SETFL, flags);
	if(flags  == -1){
		if(res == 1){
			shutdown(ct->fd, 2);
		}
		D_PRINTF(("[libnbt] fcntl error\n"));
		return LIBNBT_ERR_OTHER_ERR;
	}
#endif

	if(res < 0){
		return LIBNBT_ERR_CONNECT_ERR;
	}
	if(res == 0){
		return LIBNBT_ERR_CONNECT_TIMEOUT;
	}
		
	// tcp connection success.
	ct->state = STATE_CONNECTED_TCP;
	
	// send nbt session request
	// set called name
	res = nbt_padname(rmt->name, temp, NBT_LENGTH_NETBIOSNAME, ' ', rmt->suffix);
	if(res >= 0){
		res = nbt_enc2nd(buff, 256, temp, GSetting.scopeid);
	}
	if(res < 0){
		D_PRINTF(("[nbtsub] connect: translatefailed.\n"));
		shutdown(ct->fd, 2);
		ct->state = STATE_OPENED;
		
		return LIBNBT_ERR_INVALID_PARAM;
	}
	offset = res;
	
	// set calling name
	res = nbt_padname(local->name, temp, NBT_LENGTH_NETBIOSNAME, ' ', local->suffix);
	if(res >= 0){
		res = nbt_enc2nd(buff + offset, 256, temp, GSetting.scopeid);
	}
	if(res < 0){
		D_PRINTF(("[nbtsub] connect: translatefailed.\n"));
		shutdown(ct->fd, 2);
		ct->state = STATE_OPENED;
		
		return LIBNBT_ERR_INVALID_PARAM;
	}
	offset += res;
	
	res = inner_send(s, buff, offset, 0, NBT_MTYPE_SES_REQ);
	if(res < 0){
		shutdown(ct->fd, 2);
		ct->state = STATE_OPENED;
		
		return res;
	}
	if(res != offset){
		D_PRINTF(("[nbtsub] connect: request send failed\n"));
		shutdown(ct->fd, 2);
		ct->state = STATE_OPENED;
		
		return LIBNBT_ERR_SENDDATA_INVALID;
	}
	
	// get responce
	ct->isPacketLast = 1; /* pPbg̍Ōɂ */
	res = nbt_look(s, 0);
	if(res < 0){
		shutdown(ct->fd, 2);
		ct->state = STATE_OPENED;
		
		return res;
	}
	/* m艞ȊOȂΐؒf */
	if(res != NBT_MTYPE_SES_RES_ACK){
		// negative response
		shutdown(ct->fd, 2);
		ct->state = STATE_OPENED;
		
		return LIBNBT_ERR_RES_NEGATIVE;
	}
	
	// positive
	/* readׂ̂cĂ邩 */
	res = nbt_readable(ct->fd, 0, -1);
	if(res < 0){
		shutdown(ct->fd, 2);
		ct->state = STATE_OPENED;
		
		return LIBNBT_ERR_SELECT_ERR;
	}
	if(res == 0){
		ct->state = STATE_CONNECTED;
		return LIBNBT_ERR_NOERR;
	}
	
	// res > 0
	res = 0;
	while(res > 0){
		res = nbt_recv(s, buff, sizeof buff, 0, NBT_MSG_ONEMES);
	}
	if(res != LIBNBT_ERR_EOM){
		shutdown(ct->fd, 2);
		ct->state = STATE_OPENED;
		D_PRINTF(("[nbtsub] connect: responce receive failed\n"));
		return res;
	}
	
	ct->state = STATE_CONNECTED;
	
	return LIBNBT_ERR_NOERR;
}


/*
------------------------------------------------------------------
nbt_send()

ڑ\PbgɃf[^𑗐MB

:
IN	s:		NBTfBXNv^
IN	msg:	Mf[^i[Ăobt@̐擪AhX
IN	len:	Mf[^̒BbyteŎwB
IN	flags:	send֐ɐݒ肷tO

߂l:

:
Mf[^f[^̃oCg

s:
LIBNBT_ERR_INVALID_PARAM	(-110)
LIBNBT_ERR_SELECT_TIMEOUT	(-151)
LIBNBT_ERR_SELECT_ERR		(-150)
LIBNBT_ERR_SENDDATA_INVALID	(-500)
LIBNBT_ERR_SEND_ERR			(-170)

------------------------------------------------------------------
*/
int nbt_send(int s, const void* msg, int len, unsigned int flags)
{

	return inner_send(s, msg, len, flags, NBT_MTYPE_SES_MES);
}


/*
------------------------------------------------------------------
nbt_send_ack()

ڑ\Pbgɍm艞𑗐M


߂l:
:
LIBNBT_ERR_NOERR

s:
LIBNBT_ERR_INVALID_PARAM
LIBNBT_ERR_SELECT_TIMEOUT
LIBNBT_ERR_SELECT_ERR
LIBNBT_ERR_SENDDATA_INVALID
LIBNBT_ERR_SEND_ERR

------------------------------------------------------------------
*/
int nbt_send_ack(int s)
{
	int res;
	
	res = inner_send(s, NULL, 0, 0, NBT_MTYPE_SES_RES_ACK);
	if(res < 0){
		return res;
	}
	
	return LIBNBT_ERR_NOERR;
}


/*
------------------------------------------------------------------
nbt_send_nack()

ڑ\Pbgɔے艞𑗐M


߂l
:
LIBNBT_ERR_NOERR

s:
LIBNBT_ERR_INVALID_PARAM
LIBNBT_ERR_SELECT_TIMEOUT
LIBNBT_ERR_SELECT_ERR
LIBNBT_ERR_SENDDATA_INVALID
LIBNBT_ERR_SEND_ERR

------------------------------------------------------------------
*/
int nbt_send_nack(int s, int errortype)
{	
	int res;
	u_char buf;
	
	switch(errortype){
	case NBT_ERROR_SES_NL_CALLEDNAME: 		buf = VAL_ERROR_SES_NL_CALLEDNAME; 	break;
	case NBT_ERROR_SES_NL_CALLINGNAME: 		buf = VAL_ERROR_SES_NL_CALLINGNAME; break;
	case NBT_ERROR_SES_NP_CALLEDNAME: 		buf = VAL_ERROR_SES_NP_CALLEDNAME; 	break;
	case NBT_ERROR_SES_RES_INSUFFICIENT: 	buf = VAL_ERROR_SES_RES_INSUFFICIENT; break;
	case NBT_ERROR_SES_UNSPECIFIED: 		buf = VAL_ERROR_SES_UNSPECIFIED; 	break;
	default:
		return LIBNBT_ERR_INVALID_PARAM;
	}
	
	res = inner_send(s, &buf, 1, 0, NBT_MTYPE_SES_RES_NACK);
	if(res < 0){
		return res;
	}
	/* G[R[hƂăf[^1oCg𑗐M߁A߂l1ȊO̓G[ƂȂ */
	if(res != 1){
		return LIBNBT_ERR_SENDDATA_INVALID;
	}
	
	return LIBNBT_ERR_NOERR;
}



/*
------------------------------------------------------------------
nbt_send_change()

ڑ\PbgɐڑύXv𑗐M

:
LIBNBT_ERR_NOERR

s:
LIBNBT_ERR_INVALID_PARAM
LIBNBT_ERR_SELECT_TIMEOUT
LIBNBT_ERR_SELECT_ERR
LIBNBT_ERR_SENDDATA_INVALID
LIBNBT_ERR_SEND_ERR

------------------------------------------------------------------
*/
int nbt_send_change(int s, unsigned long naddr, unsigned short nport)
{
	int res;
#ifdef WIN32
#pragma pack(push, 1)
#endif // WIN32
	struct A{
		u_long	ipaddr;
		u_short	port;
#ifndef WIN32
	} __attribute__((__packed__));
#else
	};
#endif // WIN32

#ifdef WIN32
#pragma pack(pop)
#endif // WIN32

	u_char buf[6];
	struct A* a;
	a = (struct A*)buf;
	a->ipaddr = naddr;
	a->port = nport;

	res = inner_send(s, buf, 6, 0, NBT_MTYPE_SES_RES_CHNG);
	if(res < 0){
		return res;
	}
	if(res != 6){
		return LIBNBT_ERR_SENDDATA_INVALID;
	}
	
	return LIBNBT_ERR_NOERR;
}


/*
------------------------------------------------------------------
nbt_send_keepalive()


:
LIBNBT_ERR_NOERR

s:
LIBNBT_ERR_INVALID_PARAM
LIBNBT_ERR_SELECT_TIMEOUT
LIBNBT_ERR_SELECT_ERR
LIBNBT_ERR_SENDDATA_INVALID
LIBNBT_ERR_SEND_ERR

------------------------------------------------------------------
*/
int nbt_send_keepalive(int s)
{
	int res;
	res = inner_send(s, NULL, 0, 0, NBT_MTYPE_SES_KPALIV);
	if(res < 0){
		return res;
	}
	
	return LIBNBT_ERR_NOERR;
}


/*
------------------------------------------------------------------
nbt_recv()

ڑ\Pbgf[^󂯎

{֐R[Cxg݂ȂꍇA
{֐ōőNBT_WAIT_SEC~b̑҂̂Œ

:
IN	s:		NBTfBXNv^
OUT	buf:	Mf[^i[obt@̐擪AhX
IN	len:	̊֐̌ĂяoŎMf[^̍ő咷
IN	flags:	\Pbg֐flagsɒڐݒ肷
IN	flags2:	ptO
				NBT_MSG_ONEMES	// 1bZ[Ŵݓǂݍ

߂l:
:
	Mf[^byte
	
s:
LIBNBT_ERR_INVALID_PARAM(-110)
LIBNBT_ERR_SELECT_ERR(-150)
LIBNBT_ERR_SELECT_TIMEOUT(-151)

in readHeader_ss()...

LIBNBT_ERR_RECV_ERR(-190)
LIBNBT_ERR_RECVDATA_INVALID(-501)

LIBNBT_ERR_DISCONNECTED(-600)
LIBNBT_ERR_EOM(-700)

------------------------------------------------------------------
*/
int nbt_recv(int s, void* buf, int len, unsigned int flags, unsigned int flags2)
{
	NBTCTBL* ct;
	int res = -1;
	int curlen;
	int recvtotal;
	unsigned char* cbuf;
	int islisten;
	
	ct = getCtbl(decListNumber(s));
	if(!ct){
		/* param error */
		return LIBNBT_ERR_INVALID_PARAM;
	}
	
	/* SessionServiceOnly */
	if(ct->serviceType != NBT_STYPE_SESSION){
		return LIBNBT_ERR_INVALID_PARAM;
	}
	
	/* state check */
	if(	!(	(ct->state == STATE_CONNECTED_TCP)	||
			(ct->state == STATE_ACCEPTED_TCP)	||
			(ct->state == STATE_CONNECTED)	)	){
		
		return LIBNBT_ERR_INVALID_PARAM;
	}
		
	/* 	NBT_MSG_ONEMEStOĂ鎞A
		bZ[W̏I[ɒBĂꍇEOMԂ */
	if((flags2 & NBT_MSG_ONEMES) && ct->isPacketLast){
		return LIBNBT_ERR_EOM;
	}
	
	/*  葤ؒfĂꍇDISCONNECTEDԂ */
	if(ct->state == STATE_DISCONNECTED){
		return LIBNBT_ERR_DISCONNECTED;
	}
	
	islisten = (ct->state == STATE_ACCEPTED_TCP);
	
	cbuf = (unsigned char* )buf;
	curlen = len;
	recvtotal = 0;
	
	/* ŏɃ^CAEgȓɃCxg邱ƂmF */
	res = nbt_readable_checkalive(islisten? ct->cfd: ct->fd, NBT_WAIT_SEC);
	if(res < 0){
		return LIBNBT_ERR_SELECT_ERR;
	}
	if(res == 0){
		/* ^CAEg */
		return LIBNBT_ERR_SELECT_TIMEOUT;
	}
	if(res == 2){
		/* ؒf */
		ct->state = STATE_DISCONNECTED;
		return LIBNBT_ERR_DISCONNECTED;
	}
	
	while(curlen > 0){
		/* CxgȂꍇ邽߃^CAEg0ɐݒ */
		res = nbt_readable_checkalive(islisten? ct->cfd: ct->fd, 0);
		if(res < 0){
			return LIBNBT_ERR_SELECT_ERR;
		}
		if(res == 0){
			break;
		}
		if(res == 2){
			ct->state = STATE_DISCONNECTED;
			return LIBNBT_ERR_DISCONNECTED;
		}
		
		/* ܂wb_ǂݍłȂꍇƁAbZ[W̋Eɂꍇ̓wb_ǂ */
		if((!ct->isReadHeader) || ct->isPacketLast){
			res = readHeader_ss(ct);
			if(res < 0){
				return res;
			}
			/* EOFȂ΁AԑJڂG[Ԃ */
			if(ct->messageType == NBT_MTYPE_SES_DISCON){
				ct->state = STATE_DISCONNECTED;
				return LIBNBT_ERR_DISCONNECTED;
			}
			
			/* f[^擾邽߃Cxg҂(recvő҂ƍň҂ɂȂ邽) */
			res = nbt_readable_checkalive(islisten? ct->cfd: ct->fd, NBT_WAIT_SEC);
			if(res < 0){
				return LIBNBT_ERR_SELECT_ERR;
			}
			if(res == 0){
				break;
			}
			if(res == 2){
				ct->state = STATE_DISCONNECTED;
				return LIBNBT_ERR_DISCONNECTED;
			}	
		}
		
		/*	 obt@1bZ[W̃f[^(ǂݍݕ)傫ꍇ́A
			f[^obt@ɓǂݍނ */
		if(curlen <= (int)(ct->messageTotalLen - ct->readLen)){
			res = recv(islisten? ct->cfd: ct->fd, &cbuf[len - curlen], curlen, flags);
			if(res < 0){
				return LIBNBT_ERR_RECV_ERR;
			}
			/* EOFȂ΁AԑJڂG[Ԃ */
			if(res == 0){
				ct->state = STATE_DISCONNECTED;
				return LIBNBT_ERR_DISCONNECTED;
			}
			ct->readLen += res;
			recvtotal += res;
			curlen = 0;
		}
		/*	obt@1bZ[W̃f[^̕傫ꍇ́A
			f[^̏I[܂łobt@ɓǂݍ */
		else{
			res = recv(islisten? ct->cfd: ct->fd, &cbuf[len - curlen], ct->messageTotalLen - ct->readLen, flags);
			if(res < 0){
				return LIBNBT_ERR_RECV_ERR;
			}
			if(res == 0){
				ct->state = STATE_DISCONNECTED;
				return LIBNBT_ERR_DISCONNECTED;
			}
			curlen -= res;
			ct->readLen += res;
			recvtotal += res;
		}
		
		/*	1bZ[W̃f[^Ɠǂݍ񂾃f[^ꍇ͋EtO𗧂Ă */
		if(ct->messageTotalLen == ct->readLen){
			ct->isPacketLast = 1;
			
			if(flags2 & NBT_MSG_ONEMES){
				break;
			}
		}
	}
	
	return recvtotal;
}



/*
------------------------------------------------------------------
nbt_look()

NBTwb_ǂݎAbZ[W^Cv擾B

:
IN	s:
	nbt_open֐Ŏ擾NBTfBXNv^
	
IN	keepLastState:
	ȑȌԂێ邩ǂ\tO
	L̒lݒ肷
	
	NBT_LOOK_NOREAD(1)
	NBT_LOOK_READ(0)


߂l:
:	bZ[W^Cv

s:
LIBNBT_ERR_INVALID_PARAM
LIBNBT_ERR_ACCEPT_ERR
LIBNBT_ERR_SELECT_ERR
LIBNBT_ERR_SELECT_TIMEOUT

LIBNBT_ERR_RECV_ERR
LIBNBT_ERR_RECVDATA_INVALID

LIBNBT_ERR_DISCONNECTED

ӁF
ڑłɐ؂ĂꍇA{֐Őڑؒfꂽꍇ́A
G[(LIBNBT_ERR_DISCONNECTED)ԂB

------------------------------------------------------------------
*/
int nbt_look(int s, int keepLastState)
{
	NBTCTBL* ct;
	int res = -1;
	char buf[128];
	int islisten;
	
	ct = getCtbl(decListNumber(s));
	if(!ct){
		return LIBNBT_ERR_INVALID_PARAM;
	}
	/* ZbVT[rX */
	if(ct->serviceType != NBT_STYPE_SESSION){
		return LIBNBT_ERR_INVALID_PARAM;
	}
	
	/* Ԃ̐Ȃ */

	/* łɑ葤ؒfĂꍇ́A
	messageType̓e̕ۏ؂oȂߒڃbZ[W^CvԂ */
	if(ct->state == STATE_DISCONNECTED){
		return NBT_MTYPE_SES_DISCON;
	}
	
	/* ANBT_LOOK_NOREAD̏ꍇ́Act->messageTypeԂ */
	if(keepLastState == NBT_LOOK_NOREAD){
		return ct->messageType;
	}
	
	/* listensĂ\Pbgaccepts */
	if(ct->state == STATE_LISTENED){
		res = inner_accept(ct);
		if(res < 0){
			return res;
		}
	}
	
	islisten =
		(ct->state == STATE_LISTENED) | (ct->state == STATE_ACCEPTED_TCP);
		
	
	res = nbt_readable(islisten? ct->cfd: ct->fd, NBT_WAIT_SEC, -1);
	if(res < 0){
		return LIBNBT_ERR_SELECT_ERR;
	}
	if(res == 0){
		return LIBNBT_ERR_SELECT_TIMEOUT;
	}
			
	if((!islisten) && (!ct->isPacketLast)){
		/*	܂ȑÕbZ[Wobt@ɎcĂB
			AListenĂȂ\PbgɌċN蓾 */
		/*	ȑÕbZ[Wǂݔ΂ */
		while(res > 0){
			res = nbt_recv(s, buf, sizeof buf, 0, NBT_MSG_ONEMES);
		}
		if(res != LIBNBT_ERR_EOM){
			/* ߂lEOM(bZ[W̏I)ȊOłG[ */
			return res;
		}
	}
	
	/* NBTwb_ǂݍ݁Awb_ctɕێ */
	res = readHeader_ss(ct);
	if(res < 0){
		return res;
	}
	
	return ct->messageType;
}
