/* -*- tab-width: 4; -*- */
/*------------------------------------------------------------------*
 * Filename:		libnbt_util.c
 *------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef WIN32
#include <WinSock.h>
#include <errno.h>

#else // WIN32
#include <ctype.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/time.h>

#endif // WIN32

#include "libnbt.h"
#include "libnbt_private.h"
#include "libnbt_glob.h"
#include "libnbt_util.h"
#include "libnbt_util_private.h"
#include "libnbt_err.h"
#include "libnbt_macros.h"
#include "libnbt_debug.h"


#define KANJI


// private func def
static int dec1st(const u_char *nbtnm, u_char *nm);
static int enc1st(const char *nm, u_char *nbtnm);


/////////////
/* FOR SHIFT JIS CODE */
#ifdef KANJI
#define is_shift_jis(c) ((0x80 <= ((unsigned char) (c)) && ((unsigned char) (c)) <= 0x9f) || (0xe0 <= ((unsigned char) (c)) && ((unsigned char) (c)) <= 0xef))
#define is_shift_jis2(c) (0x40 <= ((unsigned char) (c)) && ((unsigned char) (c)) <= 0xfc && ((unsigned char) (c)) != 0x7f)
#define is_kana(c) ((0xa0 <= ((unsigned char) (c)) && ((unsigned char) (c)) <= 0xdf))
#endif	/* KANJI */


/*
------------------------------------------------------------------
strlower()

convert a string to lower case

------------------------------------------------------------------
*/
void nbt_strlower(char *s)
{
  while (*s)
    {
#ifdef KANJI
        if (is_shift_jis (*s)) {
            s += 2;
        } else if (is_kana (*s)) {
            s++;
        } else {
            if (isupper(*s))
                *s = tolower(*s);
            s++;
        }
#else
      if (isupper(*s))
          *s = tolower(*s);
      s++;
#endif /* KANJI */
    }
}



/*
------------------------------------------------------------------
strupper()

convert a string to upper case

------------------------------------------------------------------
*/
void nbt_strupper(char *s)
{
  while (*s)
    {
#ifdef KANJI
        if (is_shift_jis (*s)) {
            s += 2;
        } else if (is_kana (*s)) {
            s++;
        } else {
            if (islower(*s))
                *s = toupper(*s);
            s++;
        }
#else
      if (islower(*s))
          *s = toupper(*s);
      s++;
#endif /* KANJI */
    }
}



/*
------------------------------------------------------------------
nbt_padname()

obt@w肵Ŗ߂

߂l:
:	0
s:	-1
------------------------------------------------------------------
*/
int nbt_padname(const char* s, u_char* d, int len, unsigned char c, unsigned char suffix)
{
	int sl;
	
	if((!s) || (!d) || (len < 0)){
		return -1;
	}
	strcpy(d, s);
	nbt_strupper(d);
	
	sl = strlen(s);
	if(len >= sl + 1){
		memset(&d[sl], c, len - sl - 1);
	}
	d[len - 1] = suffix;
	
	return 0;
}



/*
------------------------------------------------------------------
nbt_devidename()

16oCgNetBIOS15oCgȓ̃k^[~l[gƕwɕ
łsnmnm̃̈悪dȂĂĂ悢


߂l:
	: 0
	s: -1

------------------------------------------------------------------
*/
int nbt_devidename(char* nm, u_char* suffix, const u_char* snm)
{
	int i;
	if((!nm) || (!suffix) || (!snm)){
		return -1;
	}
	
	*suffix = snm[NBT_LENGTH_NETBIOSNAME-1];
	for(i=NBT_LENGTH_NETBIOSNAME-2; (i>=0) && (snm[i]==' '); i--){
		;
	}
	if(i == -1){
		return -1;
	}
	memmove(nm, snm, i + 1);
	
	nm[i+1] = '\0';

	return 0;
}


/*
----------------------------------------------------------------
nbt_analyze_char()
̔


$IN	str:	R[h𔻒肷镶
IN	index:	肷镶̃CfbNX

߂l
CHAR_ASCII:			AXL[R[h
CHAR_KANA:			pJi
CHAR_SHIFT_JIS_0:	VtgJIS̑O1oCg
CHAR_SHIFT_JIS_1:	VtgJIŠ1oCg

-1:	G[

----------------------------------------------------------------
*/
int nbt_analyze_char(const char* str, int index)
{
	int i;
	int len;
	
	if(str == NULL){
		return -1;
	}
	
	len = strlen(str);
	if(len <= index){
		return -1;
	}
	
#ifdef KANJI
	for(i=0; i<len;){
		if(is_shift_jis(str[i])){
			if(index == i){
				return CHAR_SHIFT_JIS_0;
			}
			if(index == (i + 1)){
				return CHAR_SHIFT_JIS_1;
			}
			i+=2;
		}
		else if(is_kana(str[i])){
			if(index == i){
				return CHAR_KANA;
			}
			i++;
		}
		else{
			if(index == i){
				return CHAR_ASCII;
			}
			i++;
		}
	}
	
	return -1;
	
#else /* KANJI */
	return CHAR_ASCII;
#endif /* KANJI */
	
}




/*
------------------------------------------------------------------
nbt_encscopeid()

SCOPEID𕄍(encode)

:
out:		ꂽScopeID
outsize:	outwĂobt@̃TCY
in:			ScopeID, ex. NETBIOS.COM

inoutobt@łɓ삷

߂l:
	:	0
	s:	-1

Oϐ:
	Ȃ

------------------------------------------------------------------
*/
int nbt_encscopeid(unsigned char* out, int outsize, const char* in)
{
	int i;
	int total = 0;
	unsigned char nextchar;
	unsigned char temp;
	
	/* p[^`FbN */
	if(	(outsize <= 0)	||
		(out == NULL)	||
		(in == NULL)
	){
		return -1;
	}
	
	while(1){
		nextchar = in[0];
		
		/* hbgPʂɃXR[vID؂A
		؂XR[vIDinobt@outobt@Ɋi[ */
		for(i=0;
			(total < (outsize - 1))	&&
			(nextchar != 0x00)		&&
			(nextchar != '.');
			i++, total++){
			
			temp 		= nextchar;
			nextchar 	= in[i + 1];
			out[i + 1] 	= temp;
		}
		
		/* obt@eʂ`FbN */
		if(total == (outsize-1)){
			/* obt@I[o[ */
			return -1;
		}
		/* hbgPʂɋ؂XR[vhc̕outobt@̍ŏɑ */
		out[0] = (unsigned char )i;
		// memcpy(&out[1], in, i);

		/* ؂XR[vID̏I[00łꍇAI */
		if(nextchar == 0x00){
			out[i + 1] = 0x00;
			return 0;
		}
		/* hbǧɑXR[vhc͂邽߁Aobt@̃|C^ړ */
		in += (i+1);
		out += (i+1);
		total++;
	}
	// never come here
	return -1;
	
}


/*
------------------------------------------------------------------
nbt_decscopeid()

SCOPEID𕜍(decode)

:
in: ScopeID
out: ꂽScopeID
limit: inobt@\El
inoutobt@łɓ삷邱ƂmF

߂l:
	:	0
	s:	-1
------------------------------------------------------------------
*/
int nbt_decscopeid(char* out, int outsize, const unsigned char* in)
{
	int i;
	int total = 0;
	int index;
	
	/* p[^`FbN */
	if(	(outsize <= 0)	||
		(out == NULL)	||
		(in == NULL)
		){
		return -1;
	}

	/* inobt@0x00݂̂łꍇAXR[vID͋Ƃ */
	if(!*in){
		*out = '\0';
		return 0;
	}
	
	while(1){
		/* XR[vID̍ŏ̗vfɂ̓hbg܂ł̕Ă邽߁A
		̕indexɑ */
		index = (int)*in;
		
		/* hbg܂ł̕outobt@Ɋi[ */
		for(i=0;
			(total < outsize)	&&
			(i < index);
			i++, total++){
			
			out[i] = in[i + 1];
		}

		if(total == outsize){
			return -1;
		}
		// memcpy(out, &in[1], i);
		
		out += i;
		in += (i + 1);
		if(!*in){
			*out = '\0';
			return 0;
		}
		*out = '.';
		out++;
	}
	
	// never come here
	return -1;

}


/*
------------------------------------------------------------------
nbt_checkvalidscopeid()

ȃXR[vID`FbN


:
IN scopeid:	XR[vIDĂobt@̐擪AhX

߂l:
	:	0
	s:	-1
	
NULLw肵ꍇ0߂܂B
------------------------------------------------------------------
*/
int nbt_checkvalidscopeid(const char* scopeid)
{
	int i;
	int len;
	u_char before_use_dot = 0;
	
	
	if(scopeid == NULL){
		return 0;
	}
	
	len = strlen(scopeid);
	if(len == 0){
		return 0;
	}
	
	/* . n܂̂A.ŏÎ͕s */
	if((scopeid[0] == '.') || (scopeid[len-1] == '.')){
		return -1;
	}
	
	for(i=1; i<(len-1); i++){
		if(before_use_dot && (scopeid[i] == '.')){
			/* hbg */
			return -1;
		}
		
		if(scopeid[i] == '.'){
			before_use_dot = 1;
			continue;
		}
		
		before_use_dot = 0;
	}
	
	return 0;
	
}


/*
------------------------------------------------------------------
hassameval_short()

lĂ邩`FbN
lvalarrayz̗vfɑ݂邩

:
IN	val:	rl
IN	array:	rlz̐擪AhX
IN	num:	array̗vf

߂l:
	s:	-1
	:	larrayzvf̃CfbNXiԍŏɃqbǵj


------------------------------------------------------------------
*/
int hassameval_short(u_short val, u_short* array, int num)
{
	int i;
	for(i=0; array && (i<num) && (array[i]!=val); i++)
	;
	return ((!array) || (i == num))? -1: i;
}


 
/*
------------------------------------------------------------------
dec1st()

mastH[}bg疼Oɕϊ
XR[vIDΉ
nmobt@̑傫́A16oCgȏ邱

:
IN	nbtnm:	ꕄNetBIOS
OUT	nm:		16oCgNetBIOS(15 + suffix)

߂l:
	:	0
	s:	-1
	
------------------------------------------------------------------
*/
static int dec1st(const u_char *nbtnm, u_char *nm)
{
	int	i;
	int j;
	u_char c1, c2;

	j = 0;
	for(i=0; i<NBT_LENGTH_NETBIOSNAME; i++){
		c1 = nbtnm[j] - 'A';
		c2 = nbtnm[j+1] - 'A';
		if ((c1 & 0xF0) || (c2 & 0xF0)){
			return -1;
		}

		if(nm){
			nm[i] = (c1<<4) | c2;
		}
		j+=2;
	}
	return 0;
}


/*
------------------------------------------------------------------
enc1st()

OmastH[}bgɕϊ
XR[vIDΉ
nbtnmobt@̑傫́A34oCgȏ邱

:
IN	nm:		16oCgNetBIOS(15 + suffix)
OUT	nbtnm:	ꕄNetBIOS

߂l:
	:	0
	s:	-1
	
------------------------------------------------------------------
*/
static int enc1st(const char *nm, u_char *nbtnm)
{
	int i;
	char tmp[NBT_LENGTH_NETBIOSNAME];
	
	strncpy(tmp, nm, NBT_LENGTH_NETBIOSNAME);
	
	for(i=0; i<NBT_LENGTH_NETBIOSNAME; i++){
		 nbtnm[1+2*i] = 'A' + ((tmp[i]>>4)&0xF);
		 nbtnm[2+2*i] = 'A' + (tmp[i]&0xF);
	}
	nbtnm[0] = 0x20;
	nbtnm[33] = 0x00;
	
	return 0;
}


/*
------------------------------------------------------------------
nbt_enc2nd()

NetBIOS2xNetBIOSɕϊsȂ


 dst	:		ꂽNetBIOS+scopeidA
 len	:		dst|Cgobt@̃TCY
 nm		:		16ȂNetBIOS
 scope	:		XR[vID

߂l
		:	>0	(2NetBIOS+scopeid̑S)
		:	-1	s
		:	-2	s(obt@s)
------------------------------------------------------------------
*/
int nbt_enc2nd(u_char* dst, int len, const u_char* nm, const u_char* scope)
{
	int res;
	int scopelen;

	if((dst == NULL) || (nm == NULL)){
		return -1;
	}
	
	if(len < NBT_LENGTH_NETBIOSNAME_2ENC){
		return -2;
	}
	res = enc1st(nm, dst);
	if(res < 0){
		return -1;
	}
	if((scope == NULL) || (scope[0] == '\0')){
		return NBT_LENGTH_NETBIOSNAME_2ENC;
	}

	scopelen = strlen(scope);
	if((scopelen + NBT_LENGTH_NETBIOSNAME_2ENC + 1) > len){
		return -2;
	}	
	res = nbt_encscopeid(&dst[NBT_LENGTH_NETBIOSNAME_2ENC - 1],
	len - NBT_LENGTH_NETBIOSNAME_2ENC + 1, scope);
	
	if(res < 0){
		return -1;
	}

	return NBT_LENGTH_NETBIOSNAME_2ENC + scopelen + 1;
}

/*
------------------------------------------------------------------
nbt_dec2nd()

 2xNetBIOS畜sȂ


 nm			:	ꂽNetBIOS(16Œ)
 scope		:	XR[vID
 len		:	XR[vIDi[obt@̃TCY(byte)
 nbtnm		:	2xꂽNetBIOS̐擪AhX
 top		:	pPbg̐擪(NBTxł̐擪)
 readsize	:	>0	Of[^TCY(2x̃TCY)

߂l
			:	0	
			:	-1	s
			:	-2	s(XR[vIDpobt@Ȃ)
			:	-3	s(top|C^Kv)
------------------------------------------------------------------
*/
int nbt_dec2nd(u_char* nm, u_char* scope, int len, const u_char* nbtnm, const u_char* top, int* readsize,
		   u_char* uselabel, int* labelindex)
{
	int res;
	const u_char* scob;
	int totalsize = 0;
	int label = 0;

	if((nbtnm == NULL) || (len < 0)){
		return -1;
	}

	if(uselabel){
		*uselabel = 0;
	}
	if(labelindex){
		*labelindex = 0;
	}


	// unit 1
	if(nbtnm[0] == 0x20){
		// NetBIOS name
		res = dec1st(&nbtnm[1], nm);
		if(res < 0){
			return -1;
		}
		scob = &nbtnm[33];
	}
	else if(nbtnm[0] >= 0xC0){
		// Label String pointer
		// int laddr = (int)((((u_short)(nbtnm[0] - (u_char)0xC0)) << 8) | (u_short)nbtnm[1]);
		int laddr = (int)((((u_short)(nbtnm[0] & 0x3F)) << 8) | (u_short)nbtnm[1]);
		if(uselabel){ *uselabel = 1; }
		if(labelindex){ *labelindex = laddr; }
		
		if(laddr == 0){
			// null
			if(nm){
				memset(nm, 0, 16);
			}
			totalsize += 2;
			*readsize = totalsize;
			return 0;
		}
		// laddr > 0
		if(top == NULL){
			totalsize += 2;
			*readsize = totalsize;
			return -3;
		}
		res = dec1st(top + laddr + 1, nm); /* +1̂́AoCg擪ɕKĂ邩炾 */
		if(res < 0){
			return -1;
		}
		scob = top + 32 + laddr;
		label = 1;
	}
	else{
		return -1;
	}

	totalsize += 34;

	// unit2
	if(scob[0] == 0x00){
		// scopeid is none
		if(scope){
			scope[0] = '\0';
		}
		*readsize = label? 2: totalsize;
		return 0;
	}
	else{
		// scopeid exist.
		int scopelen = strlen(&scob[1]) + 1;
		int rlen = (len < scopelen)? len: scopelen;
		totalsize += scopelen;
		*readsize = label? 2: totalsize;

		if(scope){
			res = nbt_decscopeid(scope, rlen, scob);
			if(res < 0){
				return -2;
			}
		}
		return 0;
	}
	return -1;
}



/*
------------------------------------------------------------------
nbt_readable()

\Pbgǂݍ݉\ׂ


msec: ҂(MilliSeconds)
		:-1		҂
		:0		҂

߂l
	:0	CxgȂi^CAEgj
	:1	Cxg
	:-1	select֐G[(errno)
	:-2 f
	
------------------------------------------------------------------
*/
int nbt_readable(int fd, long msec, int interruptd)
{
	fd_set rset;
	struct timeval tv;
	int ret;
	int maxfd;

	if(msec < -1){
		return -1;
	}

	FD_ZERO(&rset);
	FD_SET(fd, &rset);
	if(interruptd != -1){
		FD_SET(interruptd, &rset);
		maxfd = MAX(fd, interruptd);
	}
	else{
		maxfd = fd;
	}

	if(msec != -1){
		tv.tv_sec = msec/1000;
		msec %= 1000;
		msec *= 1000;	// millisec --> microsec

		tv.tv_usec = msec;
	}

	ret = select(maxfd+1, &rset, NULL, NULL, (msec == -1)? NULL: &tv);
	if(!ret){
		return 0;
	}
	if (ret < 0){
		return -1;
	}

	if(interruptd != -1){
		if(FD_ISSET(interruptd, &rset)){
			return -2;
		}
	}
	
	return 1;
}

/*
------------------------------------------------------------------
nbt_readable_checkalive()

\Pbgǂݍ݉\ׁAڑLǂ𓯎ɒׂ
{֐́AXg[\Pbĝ݂T|[gB
̑̎ނ̃\PbgŌĂł͂ȂȂB



fd:		\PbgfBXNv^
msec:	҂(MilliSeconds)
		:-1		҂
		:0		҂

߂l
	:0	CxgȂi^CAEgj
	:1	Cxg
	:2	ڑؒfꂽ
	:-1	select֐G[(errno)

------------------------------------------------------------------
*/
int nbt_readable_checkalive(int fd, long msec)
{
	int		res;
	char	buf;
	
	res = nbt_readable(fd, msec, -1);
	if(res <= 0){
		return res;
	}
	
	/* CxgrecvāA߂l0̏ꍇ͐ڑ肪RlNVؒf */
	res = recv(fd, &buf, 1, MSG_PEEK);
	if(res < 0){
		return -1;
	}
	if(res == 0){
		return 2;
	}
	
	return 1;
}


/*
------------------------------------------------------------------
nbt_writable()

\Pbg݉\ׂ


	fd:		\PbgfBXNv^
	msec: 	҂(MilliSeconds)
			:-1		҂
			:0		҂

߂l
	:0	ݕsi^CAEgj
	:1	݉
	:-1	select֐G[(errno)
	:-2 f
------------------------------------------------------------------
*/
int nbt_writable(int fd, long msec, int interruptd)
{
	fd_set wset;
	fd_set rset;

	struct timeval tv;
	int ret;
	int maxfd;

	if(msec < -1){
		return -1;
	}
	FD_ZERO(&rset);
	FD_ZERO(&wset);
	FD_SET(fd, &wset);

	if(interruptd != -1){
		FD_SET(interruptd, &rset);
		maxfd = MAX(fd, interruptd);
	}
	else{
		maxfd = fd;
	}

	if(msec != -1){
		tv.tv_sec = msec/1000;
		msec %= 1000;
		msec *= 1000;	// millisec --> microsec

		tv.tv_usec = msec;
	}

	ret = select(maxfd+1, &rset, &wset, NULL, (msec == -1)? NULL: &tv);
	if(!ret){
		return 0;
	}
	if (ret < 0){
		return -1;
	}

	if(interruptd != -1){
		if(FD_ISSET(interruptd, &rset)){
			return -2;
		}
	}

	return 1;
}



/*
------------------------------------------------------------------
nbt_in_sendto()

f[^O\Pbgf[^𑗐M

:
	fd:			\PbgfBXNv^
	msg:		MbZ[Wobt@
	len:		MbZ[W(byte)
	flag:		sendto֐Ɏw肷tO
	dstnaddr:	MIPAhX(network byte)
	dstnport:	M|[gԍ(network byte)
	timeout:	^CAEgB~bŎw

߂l:
	:	MoCg
	s:
		LIBNBT_ERR_SENDTO_ERR		(-160)	:	sendto֐G[(errno)
		LIBNBT_ERR_SELECT_ERR		(-150)	:	select֐G[(errno)
		LIBNBT_ERR_SELECT_TIMEOUT	(-151)	:	^CAEg(Ms)
		LIBNBT_ERR_INTERRUPTED		(-2000)	:	f
		
------------------------------------------------------------------
*/
int nbt_in_sendto(int fd, const void* msg, size_t len, int flags, u_long dstnaddr, u_short dstnport,
				long timeout, int interruptd)
{
	struct sockaddr_in addr;
	int res;
	
	memset(&addr, 0, sizeof addr);
	addr.sin_family = AF_INET;
	addr.sin_port = dstnport;
	addr.sin_addr = *(struct in_addr* )&dstnaddr;
	
	res = nbt_writable(fd, timeout, interruptd);
	if(res == -2){
		return LIBNBT_ERR_INTERRUPTED;
	}
	if(res < 0){
		return LIBNBT_ERR_SELECT_ERR;
	}
	if(res == 0){
		return LIBNBT_ERR_SELECT_TIMEOUT;
	}
	
	res = sendto(fd, msg, len, flags, (struct sockaddr* )&addr, sizeof addr);
	if(res < 0){
		D_PRINTF(("inet_ntoa = %s\n", inet_ntoa(*(struct in_addr* )&dstnaddr)));
		return LIBNBT_ERR_SENDTO_ERR;
	}
	
	return res;
}


/*
------------------------------------------------------------------
nbt_in_recvfrom()

f[^O\Pbgf[^M

:
IN		fd:			\PbgfBXNv^
OUT		buf:		MbZ[Wobt@
IN		len:		MbZ[Wobt@(byte)
IN		flags:		recvfrom֐Ɏw肷tO
OUT%	rnaddr:		MIPAhX(network byte)
OUT%	rnport:		M|[gԍ(network byte)
IN		timeout:	^CAEgB~bŎw

߂l:
:
MoCg

s:
	LIBNBT_ERR_RECVFROM_ERR		(-180)	:recvfrom֐G[(errno)
	LIBNBT_ERR_SELECT_ERR		(-150)	:select֐G[(errno)
	LIBNBT_ERR_SELECT_TIMEOUT	(-151)	:^CAEg(Ms)
	LIBNBT_ERR_INTERRUPTED		(-2000)	:f

------------------------------------------------------------------
*/
int nbt_in_recvfrom(int fd, void* buf, size_t len, int flags, u_long* rnaddr, u_short* rnport,
				long timeout, int interruptd)
{
	struct sockaddr_in addr;
	int res;
	int addrlen;
	
	if(rnaddr){ *rnaddr = 0; }
	if(rnport){ *rnport = 0; }
	
	memset(&addr, 0, sizeof addr);
	addrlen = sizeof addr;
	
	res = nbt_readable(fd, timeout, interruptd);
	if(res == -2){
		return LIBNBT_ERR_INTERRUPTED;
	}

	if(res < 0){
		return LIBNBT_ERR_SELECT_ERR;
	}
	if(res == 0){
		return LIBNBT_ERR_SELECT_TIMEOUT;
	}
	
	res = recvfrom(fd, buf, len, flags, (struct sockaddr* )&addr, &addrlen);
	if(res < 0){
		return LIBNBT_ERR_RECVFROM_ERR;
	}
		
	if(rnaddr){
		*rnaddr = *(unsigned long* )&addr.sin_addr;

	}
	if(rnport){
		*rnport = addr.sin_port;
	}
	
	return res;
}

/*
------------------------------------------------------------------
nbt_in_recv()

Xg[\Pbgf[^M

:
IN	fd:			\PbgfBXNv^
OUT	buf:		Mobt@
IN	len:		Mobt@(byte)
IN	flags:		recv֐ɓntO


߂l:
:
Mf[^̃oCg

s:
LIBNBT_ERR_RECV_ERR			(-190):	recv֐G[(errno)
LIBNBT_ERR_SELECT_ERR		(-150):	select֐G[(errno)
LIBNBT_ERR_SELECT_TIMEOUT	(-151):	^CAEg(Ms)
LIBNBT_ERR_DISCONNECTED		(-600):	ؒfꂽ

------------------------------------------------------------------
*/
int nbt_in_recv(int fd, void* buf, size_t len, int flags)
{
	int res;

	res = nbt_readable_checkalive(fd, NBT_WAIT_SEC);
	if(res < 0){
		return LIBNBT_ERR_SELECT_ERR;
	}
	if(res == 2){
		return LIBNBT_ERR_SELECT_TIMEOUT;
	}
	return recv(fd, buf, len, flags);
}



/*
------------------------------------------------------------------
nbt_ignore_packet()

f[^O^\Pbgɑ΂ĂPpPbgǂݎ̂Ă

:
IN	sd:
\PbgfBXNv^

߂l:
:	
		MoCg
s:
		LIBNBT_ERR_RECVFROM_ERR		(-180):	recvfrom֐G[(errno)
		LIBNBT_ERR_SELECT_ERR		(-150):	select֐G[(errno)
		LIBNBT_ERR_SELECT_TIMEOUT	(-151):	^CAEg(Ms)
		LIBNBT_ERR_INTERRUPTED		(-2000)	:	f

------------------------------------------------------------------
*/
int nbt_ignore_packet(int sd, int interruptd)
{
	int res;
	u_char	temp[IGNORE_PACKET_BUFSIZE];
	
	res = nbt_in_recvfrom(sd, temp, sizeof temp, 0, NULL, NULL, NBT_WAIT_SEC, interruptd);
	
	return res;
}

/*
------------------------------------------------------------------
nbt_getnport()

\PbgoChĂ|[gԍ擾

:
IN	fd:	\PbgfBXNv^

߂l:
	:	|[gԍ
	s:	0

------------------------------------------------------------------
*/
unsigned short nbt_getnport(int fd)
{
	struct sockaddr_in	inaddr;
	int res;
	int namelen;
	
	namelen = sizeof inaddr;
	memset(&inaddr, 0, sizeof inaddr);
	res = getsockname(fd, (struct sockaddr* )&inaddr, &namelen);
	if(res < 0){
		return 0;
	}

	return inaddr.sin_port;
}


/*
------------------------------------------------------------------

------------------------------------------------------------------
*/
int encListNumber(int rawnum)
{
	return rawnum + NBT_FD_COUNT_BEGIN;
}


/*
------------------------------------------------------------------

------------------------------------------------------------------
*/
int decListNumber(int encnum)
{
	return encnum - NBT_FD_COUNT_BEGIN;
}


/*
------------------------------------------------------------------

------------------------------------------------------------------
*/
NBTCTBL* getCtbl(int index)
{
	if((0<= index) && (index < NBT_NBTCTBL_MAX)){
		if(NbtCtblList[index].isUsed){
			return &NbtCtblList[index];
		}
	}	
	return NULL;
}


/*
------------------------------------------------------------------
getUnusedCtbl()

gpNBTCTBL\̗̈̐擪AhX擾

:
OUT	index:	gpNBTCTBL\̗̈̔z̃CfbNX

߂l:
	:	gpNBTCTBL\̗̈̐擪AhX
	gp̈Ȃ:	NULL
	
------------------------------------------------------------------
*/
NBTCTBL* getUnusedCtbl(int* index)
{
	int i;
	
	if(index){
		*index = 0;
	}
	for(i=0; (i<NBT_NBTCTBL_MAX) && (NbtCtblList[i].isUsed!=0); i++)
	;
	if(i < NBT_NBTCTBL_MAX){
		if(index){
			*index = i;
		}
		return &NbtCtblList[i];
	}
	
	return NULL;
}


/*
------------------------------------------------------------------
Ȉmemcpy
------------------------------------------------------------------
*/
int MemcpySafe(u_char* dst, const u_char* src, int size)
{
	memmove(dst, src, size);
	return 0;
}


/*
------------------------------------------------------------------
NBTCTBL
------------------------------------------------------------------
*/
void initCtbl(NBTCTBL* ct)
{
	memset(ct, 0, sizeof(NBTCTBL));
	ct->cfd		= -1;
	ct->fd		= -1;
	ct->bcnaddr = htonl(0xFFFFFFFFUL);
}
