/*
   Unix SMB/Netbios implementation.
   Version 1.9.
   Samba utility functions
   Copyright (C) Andrew Tridgell 1992-1998
   Copyright (C) Jeremy Allison 2001

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "includes.h"

#if (defined(HAVE_NETGROUP) && defined (WITH_AUTOMOUNT))
#ifdef WITH_NISPLUS_HOME
#ifdef BROKEN_NISPLUS_INCLUDE_FILES
/*
 * The following lines are needed due to buggy include files
 * in Solaris 2.6 which define GROUP in both /usr/include/sys/acl.h and
 * also in /usr/include/rpcsvc/nis.h. The definitions conflict. JRA.
 * Also GROUP_OBJ is defined as 0x4 in /usr/include/sys/acl.h and as
 * an enum in /usr/include/rpcsvc/nis.h.
 */

#if defined(GROUP)
#undef GROUP
#endif

#if defined(GROUP_OBJ)
#undef GROUP_OBJ
#endif

#endif /* BROKEN_NISPLUS_INCLUDE_FILES */

#include <rpcsvc/nis.h>

#else /* !WITH_NISPLUS_HOME */

#if !defined(HAVE_RPCSVC_YPCLNT_H) && !defined(_nec_ews)
#include "rpcsvc/ypclnt.h"
#endif

#endif /* WITH_NISPLUS_HOME */
#endif /* HAVE_NETGROUP && WITH_AUTOMOUNT */

#ifdef WITH_SSL
#include <openssl/ssl.h>
#undef Realloc  /* SSLeay defines this and samba has a function of this name */
extern SSL  *ssl;
extern int  sslFd;
#endif  /* WITH_SSL */

int Protocol = PROTOCOL_COREPLUS;

/* a default finfo structure to ensure all fields are sensible */
file_info def_finfo = {(SMB_OFF_T)-1,(uint16)0,(time_t)0,(time_t)0,(time_t)0,(time_t)0,"",""};

/* this is used by the chaining code */
int chain_size = 0;

int trans_num = 0;

/*
   case handling on filenames 
*/
int case_default = CASE_LOWER;

/* the following control case operations - they are put here so the
   client can link easily */
BOOL case_sensitive;
BOOL case_preserve;
BOOL use_mangled_map = False;
BOOL short_case_preserve;
BOOL case_mangle;

pstring user_socket_options=DEFAULT_SOCKET_OPTIONS;   

pstring global_myname = "";
fstring global_myworkgroup = "";
char **my_netbios_names;

/****************************************************************************
 Like atoi but gets the value up to the separater character.
****************************************************************************/

char *smb_Atoic(char *p, int *n, char *c)
{
	if (!isdigit((int)*p)) {
		DEBUG_PRINT(5, ("Atoic: malformed number\n"));
		return NULL;
	}

	(*n) = atoi(p);

	while ((*p) && isdigit((int)*p))
		p++;

	if (strchr(c, *p) == NULL) {
		DEBUG_PRINT(5, ("Atoic: no separator characters (%s) not found\n", c));
		return NULL;
	}

	return p;
}

/*************************************************************************
 Reads a list of numbers.
 *************************************************************************/

char *smb_get_numlist(char *p, uint32 **num, int *count)
{
	int val;

	if (num == NULL || count == NULL)
		return NULL;

	(*count) = 0;
	(*num  ) = NULL;

	while ((p = smb_Atoic(p, &val, ":,")) != NULL && (*p) != ':') {
		uint32 *tn;

		tn = (uint32 *)smb_Realloc((*num), ((*count)+1) * sizeof(uint32));
		if (tn == NULL)
		{
			SAFE_FREE(*num);
			return NULL;
		} else
			(*num) = tn;
		(*num)[(*count)] = val;
		(*count)++;
		p++;
	}

	return p;
}

/*******************************************************************
  show a smb message structure
********************************************************************/
void smb_show_msg(char *buf)
{
	int i;
	int bcc=0;

	if (DEBUGLEVEL < 5) return;

	DEBUG_PRINT(5,("size=%d\nsmb_com=0x%x\nsmb_rcls=%d\nsmb_reh=%d\nsmb_err=%d\nsmb_flg=%d\nsmb_flg2=%d\n",
			smb_len(buf),
			(int)CVAL(buf,smb_com),
			(int)CVAL(buf,smb_rcls),
			(int)CVAL(buf,smb_reh),
			(int)SVAL(buf,smb_err),
			(int)CVAL(buf,smb_flg),
			(int)SVAL(buf,smb_flg2)));
	DEBUG_PRINT(5,("smb_tid=%d\nsmb_pid=%d\nsmb_uid=%d\nsmb_mid=%d\nsmt_wct=%d\n",
			(int)SVAL(buf,smb_tid),
			(int)SVAL(buf,smb_pid),
			(int)SVAL(buf,smb_uid),
			(int)SVAL(buf,smb_mid),
			(int)CVAL(buf,smb_wct)));

	for (i=0;i<(int)CVAL(buf,smb_wct);i++)
	{
		DEBUG_PRINT(5,("smb_vwv[%d]=%d (0x%X)\n",i,
			SVAL(buf,smb_vwv+2*i),SVAL(buf,smb_vwv+2*i)));
	}

	bcc = (int)SVAL(buf,smb_vwv+2*(CVAL(buf,smb_wct)));

	DEBUG_PRINT(5,("smb_bcc=%d\n",bcc));

	if (DEBUGLEVEL < 10) return;

	if (DEBUGLEVEL < 50)
	{
		bcc = MIN(bcc, 512);
	}

	smb_dump_data(10, smb_buf(buf), bcc);
}

/*******************************************************************
  set the length and marker of an smb packet
********************************************************************/
void smb_setlen(char *buf,int len)
{
  _smb_setlen(buf,len);

  SCVAL(buf,4,0xFF);
  SCVAL(buf,5,'S');
  SCVAL(buf,6,'M');
  SCVAL(buf,7,'B');
}

/*******************************************************************
  setup the word count and byte count for a smb message
********************************************************************/
int smb_set_message(char *buf,int num_words,int num_bytes,BOOL zero)
{
	if (zero)
		memset(buf + smb_size,'\0',num_words*2 + num_bytes);
	SCVAL(buf,smb_wct,num_words);
	SSVAL(buf,smb_vwv + num_words*SIZEOFWORD,num_bytes);  
	smb_setlen(buf,smb_size + num_words*2 + num_bytes - 4);
	return (smb_size + num_words*2 + num_bytes);
}

/*******************************************************************
  setup only the byte count for a smb message
********************************************************************/
int smb_set_message_bcc(char *buf,int num_bytes)
{
	int num_words = CVAL(buf,smb_wct);
	SSVAL(buf,smb_vwv + num_words*SIZEOFWORD,num_bytes);  
	smb_setlen(buf,smb_size + num_words*2 + num_bytes - 4);
	return (smb_size + num_words*2 + num_bytes);
}

/*******************************************************************
  setup only the byte count for a smb message, using the end of the
  message as a marker
********************************************************************/
int smb_set_message_end(void *outbuf,void *end_ptr)
{
	return smb_set_message_bcc((char *)outbuf,PTR_DIFF(end_ptr,smb_buf((char *)outbuf)));
}

/****************************************************************************
Set a fd into blocking/nonblocking mode. Uses POSIX O_NONBLOCK if available,
else
if SYSV use O_NDELAY
if BSD use FNDELAY
****************************************************************************/
int smb_set_blocking(int fd, BOOL set)
{
  int val;
#ifdef O_NONBLOCK
#define FLAG_TO_SET O_NONBLOCK
#else
#ifdef SYSV
#define FLAG_TO_SET O_NDELAY
#else /* BSD */
#define FLAG_TO_SET FNDELAY
#endif
#endif

  if((val = smb_sys_fcntl_long(fd, F_GETFL, 0)) == -1)
	return -1;
  if(set) /* Turn blocking on - ie. clear nonblock flag */
	val &= ~FLAG_TO_SET;
  else
    val |= FLAG_TO_SET;
  return smb_sys_fcntl_long( fd, F_SETFL, val);
#undef FLAG_TO_SET
}

/****************************************************************************
 Transfer some data between two fd's.
****************************************************************************/

#ifndef TRANSFER_BUF_SIZE
#define TRANSFER_BUF_SIZE 65536
#endif

ssize_t smb_transfer_file_internal(int infd, int outfd, size_t n, ssize_t (*read_fn)(int, void *, size_t),
						ssize_t (*write_fn)(int, const void *, size_t))
{
	char *buf;
	size_t total = 0;
	ssize_t read_ret;
	ssize_t write_ret;
	size_t num_to_read_thistime;
	size_t num_written = 0;

	if ((buf = (char *)smb_malloc(TRANSFER_BUF_SIZE)) == NULL)
		return -1;

	while (total < n) {
		num_to_read_thistime = MIN((n - total), TRANSFER_BUF_SIZE);

		read_ret = (*read_fn)(infd, buf, num_to_read_thistime);
		if (read_ret == -1) {
			DEBUG_PRINT(0,("transfer_file_internal: read failure. Error = %s\n", strerror(ERRNO) ));
			SAFE_FREE(buf);
			return -1;
		}
		if (read_ret == 0)
			break;

		num_written = 0;
 
		while (num_written < read_ret) {
			write_ret = (*write_fn)(outfd,buf + num_written, read_ret - num_written);
 
			if (write_ret == -1) {
				DEBUG_PRINT(0,("transfer_file_internal: write failure. Error = %s\n", strerror(ERRNO) ));
				SAFE_FREE(buf);
				return -1;
			}
			if (write_ret == 0)
				return (ssize_t)total;
 
			num_written += (size_t)write_ret;
		}

		total += (size_t)read_ret;
	}

	SAFE_FREE(buf);
	return (ssize_t)total;		
}

/*******************************************************************
 Sleep for a specified number of milliseconds.
********************************************************************/

void smb_msleep(unsigned int t)
{
	unsigned int tdiff=0;
	struct timeval tval,t1,t2;  
	fd_set fds;

	smb_GetTimeOfDay(&t1);
	smb_GetTimeOfDay(&t2);
  
	while (tdiff < t) {
		tval.tv_sec = (t-tdiff)/1000;
		tval.tv_usec = 1000*((t-tdiff)%1000);

		/* Never wait for more than 1 sec. */
		if (tval.tv_sec > 1) {
			tval.tv_sec = 1; 
			tval.tv_usec = 0;
		}

		FD_ZERO(&fds);
		ERRNO = 0;
		smb_sys_select_intr(0,&fds,NULL,NULL,&tval);

		smb_GetTimeOfDay(&t2);
		if (t2.tv_sec < t1.tv_sec) {
			/* Someone adjusted time... */
			t1 = t2;
		}

		tdiff = TvalDiff(&t1,&t2);
	}
}


/****************************************************************************
 Expand a pointer to be a particular size.
****************************************************************************/

void *smb_Realloc(void *p,size_t size)
{
	void *ret=NULL;

	if (size == 0) {
		SAFE_FREE(p);
		DEBUG_PRINT(5,("Realloc asked for 0 bytes\n"));
		return NULL;
	}

	if (!p)
		ret = (void *)smb_malloc(size);
	else
		ret = (void *)realloc(p,size);

	if (!ret)
		DEBUG_PRINT(0,("Memory allocation error: failed to expand to %d bytes\n",(int)size));

	return(ret);
}

/****************************************************************************
 Free memory, checks for NULL.
use directly SAFE_FREE()
exist only because we need to pass a function pointer somewhere --SSS
****************************************************************************/

void smb_safe_free(void *p)
{
	SAFE_FREE(p);
}

/****************************************************************************
 Interpret a protocol description string, with a default.
****************************************************************************/

int smb_interpret_protocol(char *str,int def)
{
  if (strequal(str,"NT1"))
    return(PROTOCOL_NT1);
  if (strequal(str,"LANMAN2"))
    return(PROTOCOL_LANMAN2);
  if (strequal(str,"LANMAN1"))
    return(PROTOCOL_LANMAN1);
  if (strequal(str,"CORE"))
    return(PROTOCOL_CORE);
  if (strequal(str,"COREPLUS"))
    return(PROTOCOL_COREPLUS);
  if (strequal(str,"CORE+"))
    return(PROTOCOL_COREPLUS);
  
  DEBUG_PRINT(0,("Unrecognised protocol level %s\n",str));
  
  return(def);
}

/****************************************************************************
 Return true if a string could be a pure IP address.
****************************************************************************/

BOOL smb_is_ipaddress(const char *str)
{
  BOOL pure_address = True;
  int i;
  
  for (i=0; pure_address && str[i]; i++)
    if (!(isdigit((int)str[i]) || str[i] == '.'))
      pure_address = False;

  /* Check that a pure number is not misinterpreted as an IP */
  pure_address = pure_address && (strchr(str, '.') != NULL);

  return pure_address;
}

/* 
 * @terao20040108 add
 * IPAhXϊɃlbg[NoCgI[_[
 * ϊĂȂ̂ŏC
 * @terao20031128 add
 * inet_addr݂͐̂̕w肷
 * IPAhXƂĔFĂ܂߁ASMBp
 * Cuʓrpӂ
 * {֐Kaiser3iRIFAXjŎgpт̂ł
 * {֐SMBNCAgCupłA
 * ̃W[͎gpĂ͂ȂȂ
 * (inet_addrgp邱Ɓj
 */
uint32 smb_inet_addr(const char    *name)
{
	char		*cp;
        char            str[4][4];
        u_long          parts[4];
        int             i, j;
	u_long		naddr = 0;	/* @terao20040108 */

        bzero(str, sizeof(char[4][4]));
        bzero(parts, sizeof(u_long[4]));

        /*
         * 𑬂邽1ڂdigitۂł܂fB
         * 1ڂ'.'́A肦ȂB
         */
        if (!isdigit(name[0])) {
                return(-1);
        }

        /*
         * NULL^[~l[g܂ŁAȉ̏ŃLN^𔻒fB
         * 1)digit܂́A'.'ł邱ƁB
         * 2)1̋؂digit́A1`3ł邱ƁB
         * 3)'.'ŋ؂ꂽ؂́A4邱ƁB
         */
        for (i = 0, j = 0, cp = (char *)name; ; ++cp) {
                /*
                 * digitłȂA'.'łȂ݂ꍇ́A-1^[
                 */
                if (!isdigit(*cp) && *cp != '.' && *cp) {
                        return(-1);
                }
                if (*cp == '.') {
                        /*
                         * '.' 4ȏA܂'.'̏ꍇ-1^[
                         */
                        if ((i > 2) || (j == 0)) {
                                return(-1);
                        }
                        parts[i] = atoi(&str[i][0]);
                        i++;
                        j = 0;
                }
                else if (!*cp) {
                        /*
                         * xxx.xxx.xxx.xxx `łȂ-1^[
                         */
                        if ((i != 3) || (j == 0)) {
                                return(-1);
                        }
                        parts[i] = atoi(&str[i][0]);
                        break;
                }
                else {
                        str[i][j] = *cp;
                        /*
                         *  '.''.'̊Ԃ4ȏ゠-1^[
                         */
                        if (++j > 3) {
                                return(-1);
                        }
                }
        }
        /*
         * ؂蕔255ȏłꍇ-1^[
         */
        if (parts[0] > 0xff || parts[1] > 0xff
                                || parts[2] > 0xff || parts[3] > 0xff) {
                return(-1);
        }

	/* @terao20040108 */
	naddr = (parts[0] << 24) | (parts[1] << 16) | (parts[2] << 8) | (parts[3]);

        return(htonl(naddr));
}


/****************************************************************************
interpret an internet address or name into an IP address in 4 byte form
****************************************************************************/

uint32 smb_interpret_addr(const char *str)
{
  struct hostent *hp;
  uint32 res;

  if (strcmp(str,"0.0.0.0") == 0) return(0);
  if (strcmp(str,"255.255.255.255") == 0) return(0xFFFFFFFF);

  /* if it's in the form of an IP address then get the lib to interpret it */
  if (smb_is_ipaddress(str)) {
    res = inet_addr(str);
  } else {
    /* otherwise assume it's a network name of some sort and use 
       sys_gethostbyname */
    if ((hp = (struct hostent *)smb_sys_gethostbyname(str)) == (struct hostent *)0) {
      DEBUG_PRINT(3,("sys_gethostbyname: Unknown host. %s\n",str));
      return 0;
    }
    if(hp->h_addr == NULL) {
      DEBUG_PRINT(3,("sys_gethostbyname: host address is invalid for host %s\n",str));
      return 0;
    }
    putip((char *)&res,(char *)hp->h_addr);
  }

  if (res == (uint32)-1) return(0);

  return(res);
}

/*******************************************************************
  a convenient addition to interpret_addr()
  ******************************************************************/
struct in_addr *smb_interpret_addr2(const char *str)
{
  static struct in_addr ret;
  uint32 a = smb_interpret_addr(str);
  ret.s_addr = a;
  return(&ret);
}

/*******************************************************************
 Check if an IP is the 0.0.0.0
 ******************************************************************/

BOOL smb_is_zero_ip(struct in_addr ip)
{
	uint32 a;
	putip((char *)&a,(char *)&ip);
	return(a == 0);
}

/*******************************************************************
 Set an IP to 0.0.0.0
 ******************************************************************/

void smb_zero_ip(struct in_addr *ip)
{
	static BOOL init;
	static struct in_addr ipzero;

	if (!init) {
		ipzero = *smb_interpret_addr2("0.0.0.0");
		init = True;
	}

	*ip = ipzero;
}

/*******************************************************************
are two IPs on the same subnet?
********************************************************************/
BOOL smb_same_net(struct in_addr ip1,struct in_addr ip2,struct in_addr mask)
{
  uint32 net1,net2,nmask;

  nmask = ntohl(mask.s_addr);
  net1  = ntohl(ip1.s_addr);
  net2  = ntohl(ip2.s_addr);
            
  return((net1 & nmask) == (net2 & nmask));
}



/*******************************************************************
something really nasty happened - panic!
********************************************************************/
void smb_panic(char *why)
{
	DEBUG_PRINT(0,("PANIC: %s\n", why));
	dbgflush();
	abort();
}


void smb_print_asc(int level, unsigned char *buf,int len)
{
	int i;
	for (i=0;i<len;i++)
		DEBUG_PRINT(level,("%c", isprint(buf[i])?buf[i]:'.'));
}

void smb_dump_data(int level,char *buf1,int len)
{
  unsigned char *buf = (unsigned char *)buf1;
  int i=0;
  if (len<=0) return;

  DEBUG_PRINT(level,("[%03X] ",i));
  for (i=0;i<len;) {
    DEBUG_PRINT(level,("%02X ",(int)buf[i]));
    i++;
    if (i%8 == 0) DEBUG_PRINT(level,(" "));
    if (i%16 == 0) {      
      smb_print_asc(level,&buf[i-16],8); DEBUG_PRINT(level,(" "));
      smb_print_asc(level,&buf[i-8],8); DEBUG_PRINT(level,("\n"));
      if (i<len) DEBUG_PRINT(level,("[%03X] ",i));
    }
  }
  if (i%16) {
    int n;

    n = 16 - (i%16);
    DEBUG_PRINT(level,(" "));
    if (n>8) DEBUG_PRINT(level,(" "));
    while (n--) DEBUG_PRINT(level,("   "));

    n = MIN(8,i%16);
    smb_print_asc(level,&buf[i-(i%16)],n); DEBUG_PRINT(level,(" "));
    n = (i%16) - n;
    if (n>0) smb_print_asc(level,&buf[i-n],n); 
    DEBUG_PRINT(level,("\n"));    
  }
}


/*****************************************************************
 smb_malloc that aborts with smb_panic on fail or zero size.
 *****************************************************************/  

void *smb_xmalloc(size_t size)
{
	void *p;
	if (size == 0)
		smb_panic("smb_xmalloc: called with zero size.\n");
	if ((p = (void *)smb_malloc(size)) == NULL)
		smb_panic("smb_xmalloc: malloc fail.\n");
	return p;
}

/**
 Memdup with smb_panic on fail.
**/
void *smb_xmemdup(const void *p, size_t size)
{
	void *p2;
	p2 = smb_xmalloc(size);
	memcpy(p2, p, size);
	return p2;
}

/*****************************************************************
like strdup but for memory
 *****************************************************************/  
void *smb_memdup(void *p, size_t size)
{
	void *p2;
	if (size == 0) return NULL;
	p2 = (void *) smb_malloc(size);
	if (!p2) return NULL;
	memcpy(p2, p, size);
	return p2;
}

#ifdef __INSURE__

/*******************************************************************
This routine is a trick to immediately catch errors when debugging
with insure. A xterm with a gdb is popped up when insure catches
a error. It is Linux specific.
********************************************************************/
int _smb_Insure_trap_error(int a1, int a2, int a3, int a4, int a5, int a6)
{
	static int (*fn)();
	int ret;
	char pidstr[10];
	pstring cmd = "/usr/X11R6/bin/xterm -display :0 -T Panic -n Panic -e /bin/sh -c 'cat /tmp/ierrs.*.%d ; gdb /proc/%d/exe %d'";

	slprintf(pidstr, sizeof(pidstr)-1, "%d", sys_getpid());
	pstring_sub(cmd, "%d", pidstr, False);

	if (!fn) {
		static void *h;
		h = dlopen("/usr/local/parasoft/insure++lite/lib.linux2/libinsure.so", RTLD_LAZY);
		fn = dlsym(h, "_Insure_trap_error");
	}

	ret = fn(a1, a2, a3, a4, a5, a6);

	system(cmd);

	return ret;
}
#endif

#ifdef __APPLE__ /* for Mac OS X */
BOOL smb_am_admin(void)
{
  GID_T *group_list;
  int ngroups;
  
  if((group_list = (GID_T *)smb_malloc(groups_max() * sizeof(GID_T)))==NULL) {
    DEBUG_PRINT(0,("am_admin: Malloc fail."));
    return -1;
  }

  ngroups = sys_getgroups(groups_max(), group_list);

  if (in_group( 80 /* admin */, getgid(), ngroups, group_list)) {
    smb_free(group_list);
    return True;
  } else {
    smb_free(group_list);
    return False;
  }
}

BOOL smb_is_user_name(const char *name)
{
  int i;
  int len = strlen(name);

  for(i=0;i<len; i++)
    if(!isalnum(name[i]) && (name[i]!='_')) return False;

  return True;
}

void smb_sudo_validate(const char *pass)
{
  FILE *pipe;

  if ((pipe = popen("/usr/bin/sudo -v -S >/dev/null 2>&1", "w")) == NULL)
    return;
  fprintf(pipe, "%s\n", pass);
  pclose(pipe);
}

int smb_do_sudo(const char *command, const char *pass)
{
  FILE *pipe;
  int i, status;
  char cmd_buf[1024];

  status = slprintf(cmd_buf, 1023, "/usr/bin/sudo %s >/dev/null 2>&1",command);
  if (status < 0) return -1;

  if ((pipe = popen(cmd_buf, "w"))==NULL) return -1;

  if(pass)
    fprintf(pipe, "%s\n%s\n", pass, pass);
  status = pclose(pipe);

  return WEXITSTATUS(status);
}

#endif /* __APPLE__ */
